@@ -1,8 +1,12 @@
 #' MongoDB client
 #'
 #' Connect to a MongoDB collection. Returns a [mongo] connection object with
-#' methods listed below. The [mongolite user manual](https://jeroen.github.io/mongolite/)
-#' is the best place to get started.
+#' methods listed below. Connections automatically get pooled between collection
+#' and gridfs objects to the same database.
+#'
+#' This manual page is deliberately minimal, see the
+#' [mongolite user manual](https://jeroen.github.io/mongolite/) for more details
+#' and worked examples.
 #'
 #' @export
 #' @aliases mongolite
@@ -15,8 +19,8 @@
 #' @param options additional connection options such as SSL keys/certs.
 #' @return Upon success returns a pointer to a collection on the server.
 #' The collection can be interfaced using the methods described below.
-#' @examples # Connect to mongolabs
-#' con <- mongo("mtcars", url = "mongodb://readwrite:test@ds043942.mongolab.com:43942/jeroen_test")
+#' @examples # Connect to demo server
+#' con <- mongo("mtcars", url = "mongodb://readwrite:test@mongo.opencpu.org:43942/jeroen_test")
 #' if(con$count() > 0) con$drop()
 #' con$insert(mtcars)
 #' stopifnot(con$count() == nrow(mtcars))
@@ -30,7 +34,7 @@
 #' rm(con)
 #' gc()
 #'
-#' \dontrun{
+#' \donttest{
 #' # dplyr example
 #' library(nycflights13)
 #'
@@ -84,11 +88,12 @@
 #' }
 #' @section Methods:
 #' \describe{
-#'   \item{\code{aggregate(pipeline = '{}', handler = NULL, pagesize = 1000)}}{Execute a pipeline using the Mongo aggregation framework.}
+#'   \item{\code{aggregate(pipeline = '{}', handler = NULL, pagesize = 1000, iterate = FALSE)}}{Execute a pipeline using the Mongo aggregation framework. Set \code{iterate = TRUE} to return an iterator instead of data frame.}
 #'   \item{\code{count(query = '{}')}}{Count the number of records matching a given \code{query}. Default counts all records in collection.}
+#'   \item{\code{disconnect(gc = TRUE)}}{Disconnect collection. The \emph{connection} gets disconnected once the client is not used by collections in the pool.}
 #'   \item{\code{distinct(key, query = '{}')}}{List unique values of a field given a particular query.}
 #'   \item{\code{drop()}}{Delete entire collection with all data and metadata.}
-#'   \item{\code{export(con = stdout(), bson = FALSE)}}{Streams all data from collection to a \code{\link{connection}} in \href{http://ndjson.org}{jsonlines} format (similar to \href{http://docs.mongodb.org/v2.6/reference/mongoexport/}{mongoexport}). Alternatively when \code{bson = TRUE} it outputs the binary \href{http://bsonspec.org/faq.html}{bson} format (similar to \href{http://docs.mongodb.org/manual/reference/program/mongodump/}{mongodump}).}
+#'   \item{\code{export(con = stdout(), bson = FALSE, query = '{}', fields = '{}', sort = '{"_id":1}')}}{Streams all data from collection to a \code{\link{connection}} in \href{http://ndjson.org}{jsonlines} format (similar to \href{http://docs.mongodb.org/v2.6/reference/mongoexport/}{mongoexport}). Alternatively when \code{bson = TRUE} it outputs the binary \href{http://bsonspec.org/faq.html}{bson} format (similar to \href{http://docs.mongodb.org/manual/reference/program/mongodump/}{mongodump}).}
 #'   \item{\code{find(query = '{}', fields = '{"_id" : 0}', sort = '{}', skip = 0, limit = 0, handler = NULL, pagesize = 1000)}}{Retrieve \code{fields} from records matching \code{query}. Default \code{handler} will return all data as a single dataframe.}
 #'   \item{\code{import(con, bson = FALSE)}}{Stream import data in \href{http://ndjson.org}{jsonlines} format from a \code{\link{connection}}, similar to the \href{http://docs.mongodb.org/v2.6/reference/mongoimport/}{mongoimport} utility. Alternatively when \code{bson = TRUE} it assumes the binary \href{http://bsonspec.org/faq.html}{bson} format (similar to \href{http://docs.mongodb.org/manual/reference/program/mongorestore/}{mongorestore}).}
 #'   \item{\code{index(add = NULL, remove = NULL)}}{List, add, or remove indexes from the collection. The \code{add} and \code{remove} arguments can either be a field name or json object. Returns a dataframe with current indexes.}
@@ -96,15 +101,15 @@
 #'   \item{\code{insert(data, pagesize = 1000, stop_on_error = TRUE, ...)}}{Insert rows into the collection. Argument 'data' must be a data-frame, named list (for single record) or character vector with json strings (one string for each row). For lists and data frames, arguments in \code{...} get passed to \code{\link[jsonlite:toJSON]{jsonlite::toJSON}}}
 #'   \item{\code{iterate(query = '{}', fields = '{"_id":0}', sort = '{}', skip = 0, limit = 0)}}{Runs query and returns iterator to read single records one-by-one.}
 #'   \item{\code{mapreduce(map, reduce, query = '{}', sort = '{}', limit = 0, out = NULL, scope = NULL)}}{Performs a map reduce query. The \code{map} and \code{reduce} arguments are strings containing a JavaScript function. Set \code{out} to a string to store results in a collection instead of returning.}
-#'   \item{\code{remove(query = "{}", multiple = FALSE)}}{Remove record(s) matching \code{query} from the collection.}
+#'   \item{\code{remove(query = "{}", just_one = FALSE)}}{Remove record(s) matching \code{query} from the collection.}
 #'   \item{\code{rename(name, db = NULL)}}{Change the name or database of a collection. Changing name is cheap, changing database is expensive.}
-#'   \item{\code{run(comand = '{"ping: 1}')}}{Change the name or database of a collection. Changing name is cheap, changing database is expensive.}
 #'   \item{\code{replace(query, update = '{}', upsert = FALSE)}}{Replace matching record(s) with value of the \code{update} argument.}
+#'   \item{\code{run(command = '{"ping": 1}', simplify = TRUE)}}{Run a raw mongodb command on the database. If the command returns data, output is simplified by default, but this can be disabled.}
 #'   \item{\code{update(query, update = '{"$set":{}}', upsert = FALSE, multiple = FALSE)}}{Modify fields of matching record(s) with value of the \code{update} argument.}
 #' }
 #' @references Jeroen Ooms (2014). The \code{jsonlite} Package: A Practical and Consistent Mapping Between JSON Data and \R{} Objects. \emph{arXiv:1403.2805}. \url{http://arxiv.org/abs/1403.2805}
 mongo <- function(collection = "test", db = "test", url = "mongodb://localhost", verbose = FALSE, options = ssl_options()){
-  client <- do.call(mongo_client_new, c(list(uri = url), options))
+  client <- new_client(c(list(uri = url), options))
 
   # workaround for missing 'mongoc_client_get_default_database'
   if(missing(db) || is.null(db)){
@@ -123,23 +128,17 @@ mongo <- function(collection = "test", db = "test", url = "mongodb://localhost",
   )
   if(length(options$pem_file) && file.exists(options$pem_file))
     attr(orig, "pemdata") <- readLines(options$pem_file)
-  mongo_object(col, client, verbose = verbose, orig)
+
+  rm(client) #needed for m$disconnect() to work
+  mongo_object(col, verbose = verbose, orig)
 }
 
-mongo_object <- function(col, client, verbose, orig){
+mongo_object <- function(col, verbose, orig){
   # Check if the ptr has died and automatically recreate it
   check_col <- function(){
     if(null_ptr(col)){
       message("Connection lost. Trying to reconnect with mongo...")
-      if(length(orig$options$pem_file) && !file.exists(orig$options$pem_file)){
-        orig$options$pem_file <- tempfile()
-        writeLines(attr(orig, "pemdata"), orig$options$pem_file)
-      }
-      newclient <-  do.call(mongo_client_new, c(list(uri = orig$url), orig$options))
-      newcol <- mongo_collection_new(newclient, orig$name, orig$db)
-      mongo_collection_command_simple(newcol, '{"ping":1}')
-      client <<- newclient
-      col <<- newcol
+      col <<- collection_reset(orig)
     }
   }
 
@@ -177,16 +176,15 @@ mongo_object <- function(col, client, verbose, orig){
     iterate <- function(query = '{}', fields = '{"_id":0}', sort = '{}', skip = 0, limit = 0) {
       check_col()
       cur <- mongo_collection_find(col, query = query, sort = sort, fields = fields, skip = skip, limit = limit)
-      # make sure 'col' does not go out of scope to prevent gc
-      mongo_iterator(cur, col)
+      mongo_iterator(cur)
     }
 
-    export <- function(con = stdout(), bson = FALSE){
+    export <- function(con = stdout(), bson = FALSE, query = '{}', fields = '{}', sort = '{"_id":1}'){
       check_col()
       if(isTRUE(bson)){
-        mongo_dump(col, con, verbose = verbose)
+        mongo_dump(col, con, verbose = verbose, query = query, fields = fields, sort = sort)
       } else {
-        mongo_export(col, con, verbose = verbose)
+        mongo_export(col, con, verbose = verbose, query = query, fields = fields, sort = sort)
       }
     }
 
@@ -199,10 +197,15 @@ mongo_object <- function(col, client, verbose, orig){
       }
     }
 
-    aggregate <- function(pipeline = '{}', options = '{"allowDiskUse":true}', handler = NULL, pagesize = 1000){
+    aggregate <- function(pipeline = '{}', options = '{"allowDiskUse":true}', handler = NULL,
+                          pagesize = 1000, iterate = FALSE){
       check_col()
       cur <- mongo_collection_aggregate(col, pipeline, options)
-      mongo_stream_in(cur, handler = handler, pagesize = pagesize, verbose = verbose)
+      if(isTRUE(iterate)){
+        mongo_iterator(cur)
+      } else {
+        mongo_stream_in(cur, handler = handler, pagesize = pagesize, verbose = verbose)
+      }
     }
 
     count <- function(query = '{}'){
@@ -249,11 +252,13 @@ mongo_object <- function(col, client, verbose, orig){
 
     info <- function(){
       check_col()
-      list(
-        name = mongo_collection_name(col),
+      client <- ptr_get_prot(col)
+      structure(list(
+        collection = mongo_collection_name(col),
+        db = mongo_get_default_database(client),
         stats = tryCatch(mongo_collection_stats(col), error = function(e) NULL),
-        server = mongo_client_server_status(client)
-      )
+        server = mongo_client_server_status(col)
+      ), class = "miniprint")
     }
 
     rename <- function(name, db = NULL){
@@ -267,8 +272,8 @@ mongo_object <- function(col, client, verbose, orig){
       orig
     }
 
-    run <- function(command = '{"ping: 1}'){
-      mongo_collection_command_simple(col, command)
+    run <- function(command = '{"ping": 1}', simplify = TRUE){
+      mongo_collection_command_simple(col, command, simplify)
     }
 
     index <- function(add = NULL, remove = NULL){
@@ -281,12 +286,31 @@ mongo_object <- function(col, client, verbose, orig){
 
       mongo_collection_find_indexes(col)
     }
+
+    # C driver disconnects when 'client' gets destroyed (which is protected by 'col')
+    disconnect <- function(gc = TRUE){
+      mongo_collection_disconnect(col)
+      if(isTRUE(gc))
+        base::gc()
+      invisible()
+    }
     environment()
   })
   lockEnvironment(self, TRUE)
   structure(self, class=c("mongo", "jeroen", class(self)))
 }
 
+collection_reset <- function(orig){
+  if(length(orig$options$pem_file) && !file.exists(orig$options$pem_file)){
+    orig$options$pem_file <- tempfile()
+    writeLines(attr(orig, "pemdata"), orig$options$pem_file)
+  }
+  client <- new_client(c(list(uri = orig$url), orig$options))
+  newcol <- mongo_collection_new(client, orig$name, orig$db)
+  mongo_collection_command_simple(newcol, '{"ping":1}')
+  newcol
+}
+
 #' @export
 print.mongo <- function(x, ...){
   parent.env(x)$check_col()
@@ -295,7 +319,7 @@ print.mongo <- function(x, ...){
 
 #' @export
 print.miniprint <- function(x, ...){
-  utils::str(unclass(x))
+  utils::str(unclass(x), max.level = 2)
   invisible(x)
 }
 
