@@ -25,6 +25,8 @@
 #include "mongoc-rpc-private.h"
 #include "mongoc-stream-private.h"
 #include "mongoc-server-description-private.h"
+#include "mongoc-topology-scanner-private.h"
+#include "mongoc-log.h"
 #include "utlist.h"
 
 #ifdef MONGOC_ENABLE_SSL
@@ -34,9 +36,11 @@
 #undef MONGOC_LOG_DOMAIN
 #define MONGOC_LOG_DOMAIN "async"
 
-typedef mongoc_async_cmd_result_t (*_mongoc_async_cmd_phase_t)(
+typedef mongoc_async_cmd_result_t (*_mongoc_async_cmd_phase_t) (
    mongoc_async_cmd_t *cmd);
 
+mongoc_async_cmd_result_t
+_mongoc_async_cmd_phase_initiate (mongoc_async_cmd_t *cmd);
 mongoc_async_cmd_result_t
 _mongoc_async_cmd_phase_setup (mongoc_async_cmd_t *cmd);
 mongoc_async_cmd_result_t
@@ -47,48 +51,47 @@ mongoc_async_cmd_result_t
 _mongoc_async_cmd_phase_recv_rpc (mongoc_async_cmd_t *cmd);
 
 static const _mongoc_async_cmd_phase_t gMongocCMDPhases[] = {
+   _mongoc_async_cmd_phase_initiate,
    _mongoc_async_cmd_phase_setup,
    _mongoc_async_cmd_phase_send,
    _mongoc_async_cmd_phase_recv_len,
    _mongoc_async_cmd_phase_recv_rpc,
-   NULL,  /* no callback for MONGOC_ASYNC_CMD_ERROR_STATE    */
-   NULL,  /* no callback for MONGOC_ASYNC_CMD_CANCELED_STATE */
+   NULL, /* no callback for MONGOC_ASYNC_CMD_ERROR_STATE    */
+   NULL, /* no callback for MONGOC_ASYNC_CMD_CANCELED_STATE */
 };
 
 #ifdef MONGOC_ENABLE_SSL
 int
 mongoc_async_cmd_tls_setup (mongoc_stream_t *stream,
-                            int             *events,
-                            void            *ctx,
-                            int32_t         timeout_msec,
-                            bson_error_t    *error)
+                            int *events,
+                            void *ctx,
+                            int32_t timeout_msec,
+                            bson_error_t *error)
 {
    mongoc_stream_t *tls_stream;
-   const char *host = (const char *)ctx;
+   const char *host = (const char *) ctx;
+   int retry_events = 0;
+
 
    for (tls_stream = stream; tls_stream->type != MONGOC_STREAM_TLS;
         tls_stream = mongoc_stream_get_base_stream (tls_stream)) {
    }
 
-   if (mongoc_stream_tls_do_handshake (tls_stream, timeout_msec)) {
-      if (mongoc_stream_tls_check_cert (tls_stream, host)) {
-         return 1;
-      } else {
-         bson_set_error (error, MONGOC_ERROR_STREAM,
-                         MONGOC_ERROR_STREAM_SOCKET,
-                         "Failed to verify TLS cert.");
-         return -1;
-      }
-   } else if (mongoc_stream_tls_should_retry (tls_stream)) {
-      *events = mongoc_stream_tls_should_read (tls_stream) ? POLLIN : POLLOUT;
-   } else {
-      bson_set_error (error, MONGOC_ERROR_STREAM,
-                      MONGOC_ERROR_STREAM_SOCKET,
-                      "Failed to initialize TLS state.");
-      return -1;
+#if defined(MONGOC_ENABLE_SSL_OPENSSL) || \
+   defined(MONGOC_ENABLE_SSL_SECURE_CHANNEL)
+   /* pass 0 for the timeout to begin / continue non-blocking handshake */
+   timeout_msec = 0;
+#endif
+   if (mongoc_stream_tls_handshake (
+          tls_stream, host, timeout_msec, &retry_events, error)) {
+      return 1;
    }
 
-   return 0;
+   if (retry_events) {
+      *events = retry_events;
+      return 0;
+   }
+   return -1;
 }
 #endif
 
@@ -96,9 +99,16 @@ bool
 mongoc_async_cmd_run (mongoc_async_cmd_t *acmd)
 {
    mongoc_async_cmd_result_t result;
-   int64_t rtt;
+   int64_t rtt_msec;
    _mongoc_async_cmd_phase_t phase_callback;
 
+   BSON_ASSERT (acmd);
+
+   /* if we have successfully connected to the node, call the callback. */
+   if (acmd->state == MONGOC_ASYNC_CMD_SEND) {
+      acmd->cb (acmd, MONGOC_ASYNC_CMD_CONNECTED, NULL, 0);
+   }
+
    phase_callback = gMongocCMDPhases[acmd->state];
    if (phase_callback) {
       result = phase_callback (acmd);
@@ -110,13 +120,13 @@ mongoc_async_cmd_run (mongoc_async_cmd_t *acmd)
       return true;
    }
 
-   rtt = bson_get_monotonic_time () - acmd->start_time;
+   rtt_msec = (bson_get_monotonic_time () - acmd->cmd_started) / 1000;
 
    if (result == MONGOC_ASYNC_CMD_SUCCESS) {
-      acmd->cb (result, &acmd->reply, rtt, acmd->data, &acmd->error);
+      acmd->cb (acmd, result, &acmd->reply, rtt_msec);
    } else {
       /* we're in ERROR, TIMEOUT, or CANCELED */
-      acmd->cb (result, NULL, rtt, acmd->data, &acmd->error);
+      acmd->cb (acmd, result, NULL, rtt_msec);
    }
 
    mongoc_async_cmd_destroy (acmd);
@@ -124,15 +134,14 @@ mongoc_async_cmd_run (mongoc_async_cmd_t *acmd)
 }
 
 void
-_mongoc_async_cmd_init_send (mongoc_async_cmd_t *acmd,
-                             const char         *dbname)
+_mongoc_async_cmd_init_send (mongoc_async_cmd_t *acmd, const char *dbname)
 {
    bson_snprintf (acmd->ns, sizeof acmd->ns, "%s.$cmd", dbname);
 
-   acmd->rpc.query.msg_len = 0;
-   acmd->rpc.query.request_id = ++acmd->async->request_id;
-   acmd->rpc.query.response_to = 0;
-   acmd->rpc.query.opcode = MONGOC_OPCODE_QUERY;
+   acmd->rpc.header.msg_len = 0;
+   acmd->rpc.header.request_id = ++acmd->async->request_id;
+   acmd->rpc.header.response_to = 0;
+   acmd->rpc.header.opcode = MONGOC_OPCODE_QUERY;
    acmd->rpc.query.flags = MONGOC_QUERY_SLAVE_OK;
    acmd->rpc.query.collection = acmd->ns;
    acmd->rpc.query.skip = 0;
@@ -140,16 +149,20 @@ _mongoc_async_cmd_init_send (mongoc_async_cmd_t *acmd,
    acmd->rpc.query.query = bson_get_data (&acmd->cmd);
    acmd->rpc.query.fields = NULL;
 
+   /* This will always be isMaster, which are not allowed to be compressed */
    _mongoc_rpc_gather (&acmd->rpc, &acmd->array);
-   acmd->iovec = (mongoc_iovec_t *)acmd->array.data;
+   acmd->iovec = (mongoc_iovec_t *) acmd->array.data;
    acmd->niovec = acmd->array.len;
    _mongoc_rpc_swab_to_le (&acmd->rpc);
+   acmd->bytes_written = 0;
 }
 
 void
-_mongoc_async_cmd_state_start (mongoc_async_cmd_t *acmd)
+_mongoc_async_cmd_state_start (mongoc_async_cmd_t *acmd, bool is_setup_done)
 {
-   if (acmd->setup) {
+   if (!acmd->stream) {
+      acmd->state = MONGOC_ASYNC_CMD_INITIATE;
+   } else if (acmd->setup && !is_setup_done) {
       acmd->state = MONGOC_ASYNC_CMD_SETUP;
    } else {
       acmd->state = MONGOC_ASYNC_CMD_SEND;
@@ -159,32 +172,37 @@ _mongoc_async_cmd_state_start (mongoc_async_cmd_t *acmd)
 }
 
 mongoc_async_cmd_t *
-mongoc_async_cmd_new (mongoc_async_t           *async,
-                      mongoc_stream_t          *stream,
-                      mongoc_async_cmd_setup_t  setup,
-                      void                     *setup_ctx,
-                      const char               *dbname,
-                      const bson_t             *cmd,
-                      mongoc_async_cmd_cb_t     cb,
-                      void                     *cb_data,
-                      int32_t                   timeout_msec)
+mongoc_async_cmd_new (mongoc_async_t *async,
+                      mongoc_stream_t *stream,
+                      bool is_setup_done,
+                      struct addrinfo *dns_result,
+                      mongoc_async_cmd_initiate_t initiator,
+                      int64_t initiate_delay_ms,
+                      mongoc_async_cmd_setup_t setup,
+                      void *setup_ctx,
+                      const char *dbname,
+                      const bson_t *cmd,
+                      mongoc_async_cmd_cb_t cb,
+                      void *cb_data,
+                      int64_t timeout_msec)
 {
    mongoc_async_cmd_t *acmd;
-   mongoc_async_cmd_t *tmp;
-   bool found = false;
 
    BSON_ASSERT (cmd);
    BSON_ASSERT (dbname);
-   BSON_ASSERT (stream);
 
-   acmd = (mongoc_async_cmd_t *)bson_malloc0 (sizeof (*acmd));
+   acmd = (mongoc_async_cmd_t *) bson_malloc0 (sizeof (*acmd));
    acmd->async = async;
-   acmd->expire_at = bson_get_monotonic_time () + (int64_t) timeout_msec * 1000;
+   acmd->dns_result = dns_result;
+   acmd->timeout_msec = timeout_msec;
    acmd->stream = stream;
+   acmd->initiator = initiator;
+   acmd->initiate_delay_ms = initiate_delay_ms;
    acmd->setup = setup;
    acmd->setup_ctx = setup_ctx;
    acmd->cb = cb;
    acmd->data = cb_data;
+   acmd->connect_started = bson_get_monotonic_time ();
    bson_copy_to (cmd, &acmd->cmd);
 
    _mongoc_array_init (&acmd->array, sizeof (mongoc_iovec_t));
@@ -192,24 +210,10 @@ mongoc_async_cmd_new (mongoc_async_t           *async,
 
    _mongoc_async_cmd_init_send (acmd, dbname);
 
-   _mongoc_async_cmd_state_start (acmd);
+   _mongoc_async_cmd_state_start (acmd, is_setup_done);
 
-   /* slot the cmd into the right place in the expiration list */
-   {
-      async->ncmds++;
-      DL_FOREACH (async->cmds, tmp)
-      {
-         if (tmp->expire_at >= acmd->expire_at) {
-            DL_PREPEND_ELEM (async->cmds, tmp, acmd);
-            found = true;
-            break;
-         }
-      }
-
-      if (! found) {
-         DL_APPEND (async->cmds, acmd);
-      }
-   }
+   async->ncmds++;
+   DL_APPEND (async->cmds, acmd);
 
    return acmd;
 }
@@ -235,30 +239,45 @@ mongoc_async_cmd_destroy (mongoc_async_cmd_t *acmd)
    bson_free (acmd);
 }
 
+mongoc_async_cmd_result_t
+_mongoc_async_cmd_phase_initiate (mongoc_async_cmd_t *acmd)
+{
+   acmd->stream = acmd->initiator (acmd);
+   if (!acmd->stream) {
+      return MONGOC_ASYNC_CMD_ERROR;
+   }
+   /* reset the connect started time after connection starts. */
+   acmd->connect_started = bson_get_monotonic_time ();
+   if (acmd->setup) {
+      acmd->state = MONGOC_ASYNC_CMD_SETUP;
+   } else {
+      acmd->state = MONGOC_ASYNC_CMD_SEND;
+   }
+   return MONGOC_ASYNC_CMD_IN_PROGRESS;
+}
+
 mongoc_async_cmd_result_t
 _mongoc_async_cmd_phase_setup (mongoc_async_cmd_t *acmd)
 {
-   int64_t now;
-   int64_t timeout_msec;
+   int retval;
 
-   now = bson_get_monotonic_time ();
-   timeout_msec = (acmd->expire_at - now) / 1000;
-
-   BSON_ASSERT (timeout_msec < INT32_MAX);
-
-   switch (acmd->setup (acmd->stream, &acmd->events, acmd->setup_ctx,
-                        (int32_t) timeout_msec, &acmd->error)) {
-      case -1:
-         return MONGOC_ASYNC_CMD_ERROR;
-         break;
-      case 0:
-         break;
-      case 1:
-         acmd->state = MONGOC_ASYNC_CMD_SEND;
-         acmd->events = POLLOUT;
-         break;
-      default:
-         abort();
+   BSON_ASSERT (acmd->timeout_msec < INT32_MAX);
+   retval = acmd->setup (acmd->stream,
+                         &acmd->events,
+                         acmd->setup_ctx,
+                         (int32_t) acmd->timeout_msec,
+                         &acmd->error);
+   switch (retval) {
+   case -1:
+      return MONGOC_ASYNC_CMD_ERROR;
+   case 0:
+      break;
+   case 1:
+      acmd->state = MONGOC_ASYNC_CMD_SEND;
+      acmd->events = POLLOUT;
+      break;
+   default:
+      abort ();
    }
 
    return MONGOC_ASYNC_CMD_IN_PROGRESS;
@@ -267,34 +286,69 @@ _mongoc_async_cmd_phase_setup (mongoc_async_cmd_t *acmd)
 mongoc_async_cmd_result_t
 _mongoc_async_cmd_phase_send (mongoc_async_cmd_t *acmd)
 {
+   size_t total_bytes = 0;
+   size_t offset;
    ssize_t bytes;
+   int i;
+   /* if a continued write, then iovec will be set to a temporary copy */
+   bool used_temp_iovec = false;
+   mongoc_iovec_t *iovec = acmd->iovec;
+   size_t niovec = acmd->niovec;
 
-   bytes = mongoc_stream_writev (acmd->stream, acmd->iovec, acmd->niovec, 0);
+   for (i = 0; i < acmd->niovec; i++) {
+      total_bytes += acmd->iovec[i].iov_len;
+   }
+
+   if (acmd->bytes_written > 0) {
+      BSON_ASSERT (acmd->bytes_written < total_bytes);
+      /* if bytes have been written before, compute the offset in the next
+       * iovec entry to be written. */
+      offset = acmd->bytes_written;
+
+      /* subtract the lengths of all iovec entries written so far. */
+      for (i = 0; i < acmd->niovec; i++) {
+         if (offset < acmd->iovec[i].iov_len) {
+            break;
+         }
+         offset -= acmd->iovec[i].iov_len;
+      }
+
+      BSON_ASSERT (i < acmd->niovec);
+
+      /* create a new iovec with the remaining data to be written. */
+      niovec = acmd->niovec - i;
+      iovec = bson_malloc (niovec * sizeof (mongoc_iovec_t));
+      memcpy (iovec, acmd->iovec + i, niovec * sizeof (mongoc_iovec_t));
+      iovec[0].iov_base = (char *) iovec[0].iov_base + offset;
+      iovec[0].iov_len -= offset;
+      used_temp_iovec = true;
+   }
+
+   bytes = mongoc_stream_writev (acmd->stream, iovec, niovec, 0);
+
+   if (used_temp_iovec) {
+      bson_free (iovec);
+   }
 
    if (bytes < 0) {
-      bson_set_error (&acmd->error, MONGOC_ERROR_STREAM,
+      bson_set_error (&acmd->error,
+                      MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
                       "Failed to write rpc bytes.");
       return MONGOC_ASYNC_CMD_ERROR;
    }
 
-   while (bytes) {
-      if (acmd->iovec->iov_len < (size_t)bytes) {
-         bytes -= acmd->iovec->iov_len;
-         acmd->iovec++;
-         acmd->niovec--;
-      } else {
-         acmd->iovec->iov_base = ((char *)acmd->iovec->iov_base) + bytes;
-         acmd->iovec->iov_len -= bytes;
-         bytes = 0;
-      }
+   acmd->bytes_written += bytes;
+
+   if (acmd->bytes_written < total_bytes) {
+      return MONGOC_ASYNC_CMD_IN_PROGRESS;
    }
 
    acmd->state = MONGOC_ASYNC_CMD_RECV_LEN;
    acmd->bytes_to_read = 4;
    acmd->events = POLLIN;
 
-   acmd->start_time = bson_get_monotonic_time ();
+   acmd->cmd_started = bson_get_monotonic_time ();
 
    return MONGOC_ASYNC_CMD_IN_PROGRESS;
 }
@@ -302,21 +356,21 @@ _mongoc_async_cmd_phase_send (mongoc_async_cmd_t *acmd)
 mongoc_async_cmd_result_t
 _mongoc_async_cmd_phase_recv_len (mongoc_async_cmd_t *acmd)
 {
-   ssize_t bytes = _mongoc_buffer_try_append_from_stream (&acmd->buffer,
-                                                          acmd->stream,
-                                                          acmd->bytes_to_read,
-                                                          0, &acmd->error);
+   ssize_t bytes = _mongoc_buffer_try_append_from_stream (
+      &acmd->buffer, acmd->stream, acmd->bytes_to_read, 0);
    uint32_t msg_len;
 
    if (bytes < 0) {
-      bson_set_error (&acmd->error, MONGOC_ERROR_STREAM,
+      bson_set_error (&acmd->error,
+                      MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
                       "Failed to receive length header from server.");
       return MONGOC_ASYNC_CMD_ERROR;
    }
 
    if (bytes == 0) {
-      bson_set_error (&acmd->error, MONGOC_ERROR_STREAM,
+      bson_set_error (&acmd->error,
+                      MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
                       "Server closed connection.");
       return MONGOC_ASYNC_CMD_ERROR;
@@ -329,7 +383,8 @@ _mongoc_async_cmd_phase_recv_len (mongoc_async_cmd_t *acmd)
       msg_len = BSON_UINT32_FROM_LE (msg_len);
 
       if ((msg_len < 16) || (msg_len > MONGOC_DEFAULT_MAX_MSG_SIZE)) {
-         bson_set_error (&acmd->error, MONGOC_ERROR_PROTOCOL,
+         bson_set_error (&acmd->error,
+                         MONGOC_ERROR_PROTOCOL,
                          MONGOC_ERROR_PROTOCOL_INVALID_REPLY,
                          "Invalid reply from server.");
          return MONGOC_ASYNC_CMD_ERROR;
@@ -347,20 +402,20 @@ _mongoc_async_cmd_phase_recv_len (mongoc_async_cmd_t *acmd)
 mongoc_async_cmd_result_t
 _mongoc_async_cmd_phase_recv_rpc (mongoc_async_cmd_t *acmd)
 {
-   ssize_t bytes = _mongoc_buffer_try_append_from_stream (&acmd->buffer,
-                                                          acmd->stream,
-                                                          acmd->bytes_to_read,
-                                                          0, &acmd->error);
+   ssize_t bytes = _mongoc_buffer_try_append_from_stream (
+      &acmd->buffer, acmd->stream, acmd->bytes_to_read, 0);
 
    if (bytes < 0) {
-      bson_set_error (&acmd->error, MONGOC_ERROR_STREAM,
+      bson_set_error (&acmd->error,
+                      MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
                       "Failed to receive rpc bytes from server.");
       return MONGOC_ASYNC_CMD_ERROR;
    }
 
    if (bytes == 0) {
-      bson_set_error (&acmd->error, MONGOC_ERROR_STREAM,
+      bson_set_error (&acmd->error,
+                      MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
                       "Server closed connection.");
       return MONGOC_ASYNC_CMD_ERROR;
@@ -369,30 +424,42 @@ _mongoc_async_cmd_phase_recv_rpc (mongoc_async_cmd_t *acmd)
    acmd->bytes_to_read -= bytes;
 
    if (!acmd->bytes_to_read) {
-      if (!_mongoc_rpc_scatter (&acmd->rpc, acmd->buffer.data,
-                                acmd->buffer.len)) {
+      if (!_mongoc_rpc_scatter (
+             &acmd->rpc, acmd->buffer.data, acmd->buffer.len)) {
          bson_set_error (&acmd->error,
                          MONGOC_ERROR_PROTOCOL,
                          MONGOC_ERROR_PROTOCOL_INVALID_REPLY,
                          "Invalid reply from server.");
          return MONGOC_ASYNC_CMD_ERROR;
       }
+      if (BSON_UINT32_FROM_LE (acmd->rpc.header.opcode) ==
+          MONGOC_OPCODE_COMPRESSED) {
+         uint8_t *buf = NULL;
+         size_t len =
+            BSON_UINT32_FROM_LE (acmd->rpc.compressed.uncompressed_size) +
+            sizeof (mongoc_rpc_header_t);
+
+         buf = bson_malloc0 (len);
+         if (!_mongoc_rpc_decompress (&acmd->rpc, buf, len)) {
+            bson_free (buf);
+            bson_set_error (&acmd->error,
+                            MONGOC_ERROR_PROTOCOL,
+                            MONGOC_ERROR_PROTOCOL_INVALID_REPLY,
+                            "Could not decompress server reply");
+            return MONGOC_ASYNC_CMD_ERROR;
+         }
+
+         _mongoc_buffer_destroy (&acmd->buffer);
+         _mongoc_buffer_init (&acmd->buffer, buf, len, NULL, NULL);
+      }
 
       _mongoc_rpc_swab_from_le (&acmd->rpc);
 
-      if (acmd->rpc.header.opcode != MONGOC_OPCODE_REPLY) {
+      if (!_mongoc_rpc_get_first_document (&acmd->rpc, &acmd->reply)) {
          bson_set_error (&acmd->error,
                          MONGOC_ERROR_PROTOCOL,
                          MONGOC_ERROR_PROTOCOL_INVALID_REPLY,
-                         "Invalid reply from server.");
-         return MONGOC_ASYNC_CMD_ERROR;
-      }
-
-      if (!_mongoc_rpc_reply_get_first (&acmd->rpc.reply, &acmd->reply)) {
-         bson_set_error (&acmd->error,
-                         MONGOC_ERROR_BSON,
-                         MONGOC_ERROR_BSON_INVALID,
-                         "Failed to decode reply BSON document.");
+                         "Invalid reply from server");
          return MONGOC_ASYNC_CMD_ERROR;
       }
 
