@@ -21,6 +21,8 @@
 #include "mongoc-trace-private.h"
 #include "mongoc-write-concern-private.h"
 #include "mongoc-util-private.h"
+#include "mongoc-opts-private.h"
+#include "mongoc-write-command-private.h"
 
 
 /*
@@ -104,157 +106,153 @@ mongoc_bulk_operation_destroy (mongoc_bulk_operation_t *bulk) /* IN */
       mongoc_write_concern_destroy (bulk->write_concern);
       _mongoc_array_destroy (&bulk->commands);
 
-      if (bulk->executed) {
-         _mongoc_write_result_destroy (&bulk->result);
-      }
+      _mongoc_write_result_destroy (&bulk->result);
 
       bson_free (bulk);
    }
 }
 
 
-/* for speed, pre-split batch every 1000 docs. a future server's
- * maxWriteBatchSize may grow larger than the default, then we'll revise. */
-#define SHOULD_APPEND(_write_cmd, _write_cmd_type) \
-   (((_write_cmd->type) == (_write_cmd_type)) &&   \
-    (_write_cmd)->n_documents < MONGOC_DEFAULT_WRITE_BATCH_SIZE)
-
 /* already failed, e.g. a bad call to mongoc_bulk_operation_insert? */
-#define BULK_EXIT_IF_PRIOR_ERROR \
-   do { \
+#define BULK_EXIT_IF_PRIOR_ERROR       \
+   do {                                \
       if (bulk->result.error.domain) { \
-         EXIT; \
-      } \
+         EXIT;                         \
+      }                                \
    } while (0)
 
-#define BULK_RETURN_IF_PRIOR_ERROR \
-   do { \
-      if (bulk->result.error.domain) {\
-         if (error != &bulk->result.error) {\
-            bson_set_error (error, \
-               MONGOC_ERROR_COMMAND, \
-               MONGOC_ERROR_COMMAND_INVALID_ARG, \
-               "Bulk operation is invalid from prior error: %s", \
-               bulk->result.error.message); \
-         }; \
-         return false; \
-      }; \
+#define BULK_RETURN_IF_PRIOR_ERROR                                            \
+   do {                                                                       \
+      if (bulk->result.error.domain) {                                        \
+         if (error != &bulk->result.error) {                                  \
+            bson_set_error (error,                                            \
+                            MONGOC_ERROR_COMMAND,                             \
+                            MONGOC_ERROR_COMMAND_INVALID_ARG,                 \
+                            "Bulk operation is invalid from prior error: %s", \
+                            bulk->result.error.message);                      \
+         };                                                                   \
+         return false;                                                        \
+      };                                                                      \
    } while (0)
 
 
 bool
-_mongoc_bulk_operation_remove_with_opts (mongoc_bulk_operation_t *bulk,
-                                         const bson_t *selector,
-                                         const bson_t *opts,
-                                         bson_error_t *error) /* OUT */
+_mongoc_bulk_operation_remove_with_opts (
+   mongoc_bulk_operation_t *bulk,
+   const bson_t *selector,
+   const mongoc_bulk_remove_opts_t *remove_opts,
+   int32_t limit,
+   bson_error_t *error) /* OUT */
 {
    mongoc_write_command_t command = {0};
    mongoc_write_command_t *last;
+   bson_t opts;
+   bool has_collation;
+   bool ret = false;
 
    ENTRY;
 
    BSON_ASSERT (bulk);
    BSON_ASSERT (selector);
 
-   BULK_RETURN_IF_PRIOR_ERROR;
+   bson_init (&opts);
+
+   /* allow "limit" in opts, but it must be the correct limit */
+   if (remove_opts->limit != limit) {
+      bson_set_error (error,
+                      MONGOC_ERROR_COMMAND,
+                      MONGOC_ERROR_COMMAND_INVALID_ARG,
+                      "Invalid \"limit\" in opts: %" PRId32 "."
+                      " The value must be %" PRId32 ", or omitted.",
+                      remove_opts->limit,
+                      limit);
+      GOTO (done);
+   }
+
+   bson_append_int32 (&opts, "limit", 5, limit);
+   has_collation = !bson_empty (&remove_opts->collation);
+   if (has_collation) {
+      bson_append_document (&opts, "collation", 9, &remove_opts->collation);
+   }
 
    if (bulk->commands.len) {
       last = &_mongoc_array_index (
          &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
-      if (SHOULD_APPEND (last, MONGOC_WRITE_COMMAND_DELETE)) {
-         _mongoc_write_command_delete_append (last, selector, opts);
-         RETURN (true);
+      if (last->type == MONGOC_WRITE_COMMAND_DELETE) {
+         last->flags.has_collation |= has_collation;
+         last->flags.has_multi_write |= (remove_opts->limit == 0);
+         _mongoc_write_command_delete_append (last, selector, &opts);
+         ret = true;
+         GOTO (done);
       }
    }
 
    _mongoc_write_command_init_delete (
-      &command, selector, opts, bulk->flags, bulk->operation_id);
+      &command, selector, NULL, &opts, bulk->flags, bulk->operation_id);
+
+   command.flags.has_collation = has_collation;
+   command.flags.has_multi_write = (remove_opts->limit == 0);
 
    _mongoc_array_append_val (&bulk->commands, command);
+   ret = true;
 
-   RETURN (true);
+done:
+   bson_destroy (&opts);
+   RETURN (ret);
 }
 
+
 bool
 mongoc_bulk_operation_remove_one_with_opts (mongoc_bulk_operation_t *bulk,
                                             const bson_t *selector,
                                             const bson_t *opts,
                                             bson_error_t *error) /* OUT */
 {
-   bool retval;
-   bson_t opts_dup;
-   bson_iter_t iter;
+   mongoc_bulk_remove_one_opts_t remove_opts;
+   bool ret;
 
    ENTRY;
 
    BULK_RETURN_IF_PRIOR_ERROR;
 
-   if (opts && bson_iter_init_find (&iter, opts, "limit")) {
-      if ((!BSON_ITER_HOLDS_INT32 (&iter) && !BSON_ITER_HOLDS_INT64 (&iter)) ||
-          !bson_iter_as_int64 (&iter)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_COMMAND_INVALID_ARG,
-                         "%s expects the 'limit' option to be 1",
-                         BSON_FUNC);
-         RETURN (false);
-      }
-
-      return _mongoc_bulk_operation_remove_with_opts (
-         bulk, selector, opts, error);
+   if (!_mongoc_bulk_remove_one_opts_parse (
+          bulk->client, opts, &remove_opts, error)) {
+      _mongoc_bulk_remove_one_opts_cleanup (&remove_opts);
+      RETURN (false);
    }
 
-   bson_init (&opts_dup);
-   BSON_APPEND_INT32 (&opts_dup, "limit", 1);
-   if (opts) {
-      bson_concat (&opts_dup, opts);
-   }
-   retval = _mongoc_bulk_operation_remove_with_opts (
-      bulk, selector, &opts_dup, error);
-   bson_destroy (&opts_dup);
+   ret = _mongoc_bulk_operation_remove_with_opts (
+      bulk, selector, &remove_opts.remove, 1, error);
 
-   RETURN (retval);
+   _mongoc_bulk_remove_one_opts_cleanup (&remove_opts);
+   RETURN (ret);
 }
 
+
 bool
 mongoc_bulk_operation_remove_many_with_opts (mongoc_bulk_operation_t *bulk,
                                              const bson_t *selector,
                                              const bson_t *opts,
                                              bson_error_t *error) /* OUT */
 {
-   bool retval;
-   bson_t opts_dup;
-   bson_iter_t iter;
+   mongoc_bulk_remove_many_opts_t remove_opts;
+   bool ret;
 
    ENTRY;
 
    BULK_RETURN_IF_PRIOR_ERROR;
 
-   if (opts && bson_iter_init_find (&iter, opts, "limit")) {
-      if ((!BSON_ITER_HOLDS_INT32 (&iter) && !BSON_ITER_HOLDS_INT64 (&iter)) ||
-          bson_iter_as_int64 (&iter)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_COMMAND_INVALID_ARG,
-                         "%s expects the 'limit' option to be 0",
-                         BSON_FUNC);
-         RETURN (false);
-      }
-
-      RETURN (_mongoc_bulk_operation_remove_with_opts (
-         bulk, selector, opts, error));
+   if (!_mongoc_bulk_remove_many_opts_parse (
+          bulk->client, opts, &remove_opts, error)) {
+      _mongoc_bulk_remove_many_opts_cleanup (&remove_opts);
+      RETURN (false);
    }
 
-   bson_init (&opts_dup);
-   BSON_APPEND_INT32 (&opts_dup, "limit", 0);
-   if (opts) {
-      bson_concat (&opts_dup, opts);
-   }
-   retval = _mongoc_bulk_operation_remove_with_opts (
-      bulk, selector, &opts_dup, error);
-   bson_destroy (&opts_dup);
+   ret = _mongoc_bulk_operation_remove_with_opts (
+      bulk, selector, &remove_opts.remove, 0, error);
 
-   RETURN (retval);
+   _mongoc_bulk_remove_many_opts_cleanup (&remove_opts);
+   RETURN (ret);
 }
 
 
@@ -262,20 +260,16 @@ void
 mongoc_bulk_operation_remove (mongoc_bulk_operation_t *bulk, /* IN */
                               const bson_t *selector)        /* IN */
 {
-   bson_t opts;
    bson_error_t *error = &bulk->result.error;
 
    ENTRY;
 
    BULK_EXIT_IF_PRIOR_ERROR;
 
-   bson_init (&opts);
-   BSON_APPEND_INT32 (&opts, "limit", 0);
-
-   mongoc_bulk_operation_remove_many_with_opts (
-      bulk, selector, &opts, error);
-
-   bson_destroy (&opts);
+   if (!mongoc_bulk_operation_remove_many_with_opts (
+         bulk, selector, NULL, error)) {
+      MONGOC_WARNING ("%s", error->message);
+   }
 
    if (error->domain) {
       MONGOC_WARNING ("%s", error->message);
@@ -289,19 +283,16 @@ void
 mongoc_bulk_operation_remove_one (mongoc_bulk_operation_t *bulk, /* IN */
                                   const bson_t *selector)        /* IN */
 {
-   bson_t opts;
    bson_error_t *error = &bulk->result.error;
 
    ENTRY;
 
    BULK_EXIT_IF_PRIOR_ERROR;
 
-   bson_init (&opts);
-   BSON_APPEND_INT32 (&opts, "limit", 1);
-
-   mongoc_bulk_operation_remove_one_with_opts (
-      bulk, selector, &opts, error);
-   bson_destroy (&opts);
+   if (!mongoc_bulk_operation_remove_one_with_opts (
+         bulk, selector, NULL, error)) {
+      MONGOC_WARNING ("%s", error->message);
+   }
 
    if (error->domain) {
       MONGOC_WARNING ("%s", error->message);
@@ -336,184 +327,165 @@ void
 mongoc_bulk_operation_insert (mongoc_bulk_operation_t *bulk,
                               const bson_t *document)
 {
+   ENTRY;
+
+   BSON_ASSERT (bulk);
+   BSON_ASSERT (document);
+
+   if (!mongoc_bulk_operation_insert_with_opts (
+          bulk, document, NULL /* opts */, &bulk->result.error)) {
+      MONGOC_WARNING ("%s", bulk->result.error.message);
+   }
+
+   EXIT;
+}
+
+bool
+mongoc_bulk_operation_insert_with_opts (mongoc_bulk_operation_t *bulk,
+                                        const bson_t *document,
+                                        const bson_t *opts,
+                                        bson_error_t *error)
+{
+   mongoc_bulk_insert_opts_t insert_opts;
    mongoc_write_command_t command = {0};
    mongoc_write_command_t *last;
+   bool ret = false;
 
    ENTRY;
 
    BSON_ASSERT (bulk);
    BSON_ASSERT (document);
 
-   BULK_EXIT_IF_PRIOR_ERROR;
+   BULK_RETURN_IF_PRIOR_ERROR;
 
-   if (!_mongoc_validate_new_document (document, &bulk->result.error)) {
-      MONGOC_WARNING ("%s", bulk->result.error.message);
-      EXIT;
+   if (!_mongoc_bulk_insert_opts_parse (
+          bulk->client, opts, &insert_opts, error)) {
+      GOTO (done);
+   }
+
+   if (!_mongoc_validate_new_document (document, insert_opts.validate, error)) {
+      GOTO (done);
    }
 
    if (bulk->commands.len) {
       last = &_mongoc_array_index (
          &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
 
-      if (SHOULD_APPEND (last, MONGOC_WRITE_COMMAND_INSERT)) {
+      if (last->type == MONGOC_WRITE_COMMAND_INSERT) {
          _mongoc_write_command_insert_append (last, document);
-         EXIT;
+         ret = true;
+         GOTO (done);
       }
    }
 
    _mongoc_write_command_init_insert (
       &command,
       document,
+      opts,
       bulk->flags,
       bulk->operation_id,
       !mongoc_write_concern_is_acknowledged (bulk->write_concern));
 
    _mongoc_array_append_val (&bulk->commands, command);
 
-   EXIT;
+   ret = true;
+
+done:
+   _mongoc_bulk_insert_opts_cleanup (&insert_opts);
+   RETURN (ret);
 }
 
-bool
-_mongoc_bulk_operation_replace_one_with_opts (mongoc_bulk_operation_t *bulk,
-                                              const bson_t *selector,
-                                              const bson_t *document,
-                                              const bson_t *opts,
-                                              bson_error_t *error) /* OUT */
+static void
+_mongoc_bulk_operation_update_append (
+   mongoc_bulk_operation_t *bulk,
+   const bson_t *selector,
+   const bson_t *document,
+   const mongoc_bulk_update_opts_t *update_opts,
+   const bson_t *array_filters,
+   const bson_t *extra_opts)
 {
    mongoc_write_command_t command = {0};
    mongoc_write_command_t *last;
-
-   ENTRY;
-
-   BULK_RETURN_IF_PRIOR_ERROR;
-
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (selector);
-   BSON_ASSERT (document);
-
-   if (!_mongoc_validate_replace (document, error)) {
-      RETURN (false);
-   }
-
-   if (bulk->commands.len) {
-      last = &_mongoc_array_index (
-         &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
-      if (SHOULD_APPEND (last, MONGOC_WRITE_COMMAND_UPDATE)) {
-         _mongoc_write_command_update_append (last, selector, document, opts);
-         RETURN (true);
-      }
-   }
-
-   _mongoc_write_command_init_update (
-      &command, selector, document, opts, bulk->flags, bulk->operation_id);
-   _mongoc_array_append_val (&bulk->commands, command);
-
-   RETURN (true);
-}
-
-void
-mongoc_bulk_operation_replace_one (mongoc_bulk_operation_t *bulk,
-                                   const bson_t *selector,
-                                   const bson_t *document,
-                                   bool upsert)
-{
    bson_t opts;
-   bson_error_t *error = &bulk->result.error;
-
-   ENTRY;
+   bool has_collation;
+   bool has_array_filters;
 
    bson_init (&opts);
-   BSON_APPEND_BOOL (&opts, "upsert", upsert);
-   BSON_APPEND_BOOL (&opts, "multi", false);
+   bson_append_bool (&opts, "upsert", 6, update_opts->upsert);
+   bson_append_bool (&opts, "multi", 5, update_opts->multi);
 
-   _mongoc_bulk_operation_replace_one_with_opts (
-      bulk, selector, document, &opts, error);
-   bson_destroy (&opts);
-
-   if (error->domain) {
-      MONGOC_WARNING ("%s", error->message);
+   has_array_filters = !bson_empty0 (array_filters);
+   if (has_array_filters) {
+      bson_append_document (&opts, "arrayFilters", 12, array_filters);
    }
 
-   EXIT;
-}
-
-bool
-mongoc_bulk_operation_replace_one_with_opts (mongoc_bulk_operation_t *bulk,
-                                             const bson_t *selector,
-                                             const bson_t *document,
-                                             const bson_t *opts,
-                                             bson_error_t *error) /* OUT */
-{
-   bson_iter_t iter;
-   bson_t opts_dup;
-   bool retval;
-
-   ENTRY;
-
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (selector);
-   BSON_ASSERT (document);
-
-   if (opts && bson_iter_init_find (&iter, opts, "multi")) {
-      if (!BSON_ITER_HOLDS_BOOL (&iter) || bson_iter_bool (&iter)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_COMMAND_INVALID_ARG,
-                         "%s expects the 'multi' option to be false",
-                         BSON_FUNC);
-         RETURN (false);
-      }
-      retval = _mongoc_bulk_operation_replace_one_with_opts (
-         bulk, selector, document, opts, error);
-   } else {
-      bson_init (&opts_dup);
-      BSON_APPEND_BOOL (&opts_dup, "multi", false);
-
-      if (opts) {
-         bson_concat (&opts_dup, opts);
-      }
-      retval = _mongoc_bulk_operation_replace_one_with_opts (
-         bulk, selector, document, &opts_dup, error);
-      bson_destroy (&opts_dup);
+   has_collation = !bson_empty (&update_opts->collation);
+   if (has_collation) {
+      bson_append_document (&opts, "collation", 9, &update_opts->collation);
    }
 
-   RETURN (retval);
-}
-
-bool
-_mongoc_bulk_operation_update_with_opts (mongoc_bulk_operation_t *bulk,
-                                         const bson_t *selector,
-                                         const bson_t *document,
-                                         const bson_t *opts,
-                                         bson_error_t *error) /* OUT */
-{
-   mongoc_write_command_t command = {0};
-   mongoc_write_command_t *last;
-
-   ENTRY;
-
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (selector);
-   BSON_ASSERT (document);
-
-   BULK_RETURN_IF_PRIOR_ERROR;
-
-   if (!_mongoc_validate_update (document, error)) {
-      RETURN (false);
+   if (extra_opts) {
+      bson_concat (&opts, extra_opts);
    }
 
    if (bulk->commands.len) {
       last = &_mongoc_array_index (
          &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
-      if (SHOULD_APPEND (last, MONGOC_WRITE_COMMAND_UPDATE)) {
-         _mongoc_write_command_update_append (last, selector, document, opts);
-         RETURN (true);
+      if (last->type == MONGOC_WRITE_COMMAND_UPDATE) {
+         last->flags.has_collation |= has_collation;
+         last->flags.has_multi_write |= update_opts->multi;
+         _mongoc_write_command_update_append (last, selector, document, &opts);
+         bson_destroy (&opts);
+         return;
       }
    }
 
    _mongoc_write_command_init_update (
-      &command, selector, document, opts, bulk->flags, bulk->operation_id);
+      &command, selector, document, &opts, bulk->flags, bulk->operation_id);
+
+   command.flags.has_array_filters = has_array_filters;
+   command.flags.has_collation = has_collation;
+   command.flags.has_multi_write = update_opts->multi;
+
    _mongoc_array_append_val (&bulk->commands, command);
+   bson_destroy (&opts);
+}
+
+static bool
+_mongoc_bulk_operation_update_with_opts (
+   mongoc_bulk_operation_t *bulk,
+   const bson_t *selector,
+   const bson_t *document,
+   const mongoc_bulk_update_opts_t *update_opts,
+   const bson_t *array_filters,
+   const bson_t *extra_opts,
+   bool multi,
+   bson_error_t *error) /* OUT */
+{
+   ENTRY;
+
+   BSON_ASSERT (bulk);
+   BSON_ASSERT (selector);
+   BSON_ASSERT (document);
+
+   if (!_mongoc_validate_update (document, update_opts->validate, error)) {
+      RETURN (false);
+   }
+
+   /* allow "multi" in opts, but it must be the correct multi */
+   if (update_opts->multi != multi) {
+      bson_set_error (error,
+                      MONGOC_ERROR_COMMAND,
+                      MONGOC_ERROR_COMMAND_INVALID_ARG,
+                      "Invalid \"multi\" in opts: %s."
+                      " The value must be %s, or omitted.",
+                      update_opts->multi ? "true" : "false",
+                      multi ? "true" : "false");
+      RETURN (false);
+   }
+
+   _mongoc_bulk_operation_update_append (
+      bulk, selector, document, update_opts, array_filters, extra_opts);
 
    RETURN (true);
 }
@@ -525,36 +497,30 @@ mongoc_bulk_operation_update_one_with_opts (mongoc_bulk_operation_t *bulk,
                                             const bson_t *opts,
                                             bson_error_t *error) /* OUT */
 {
-   bool retval;
-   bson_t opts_dup;
-   bson_iter_t iter;
+   mongoc_bulk_update_one_opts_t update_opts;
+   bool ret;
 
    ENTRY;
 
-   if (opts && bson_iter_init_find (&iter, opts, "multi")) {
-      if (!BSON_ITER_HOLDS_BOOL (&iter) || bson_iter_bool (&iter)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_COMMAND_INVALID_ARG,
-                         "%s expects the 'multi' option to be false",
-                         BSON_FUNC);
-         RETURN (false);
-      }
+   BULK_RETURN_IF_PRIOR_ERROR;
 
-      RETURN (_mongoc_bulk_operation_update_with_opts (
-         bulk, selector, document, opts, error));
+   if (!_mongoc_bulk_update_one_opts_parse (
+          bulk->client, opts, &update_opts, error)) {
+      _mongoc_bulk_update_one_opts_cleanup (&update_opts);
+      RETURN (false);
    }
 
-   bson_init (&opts_dup);
-   BSON_APPEND_BOOL (&opts_dup, "multi", false);
-   if (opts) {
-      bson_concat (&opts_dup, opts);
-   }
-   retval = _mongoc_bulk_operation_update_with_opts (
-      bulk, selector, document, &opts_dup, error);
-   bson_destroy (&opts_dup);
+   ret = _mongoc_bulk_operation_update_with_opts (bulk,
+                                                  selector,
+                                                  document,
+                                                  &update_opts.update,
+                                                  &update_opts.arrayFilters,
+                                                  &update_opts.extra,
+                                                  false /* multi */,
+                                                  error);
 
-   RETURN (retval);
+   _mongoc_bulk_update_one_opts_cleanup (&update_opts);
+   RETURN (ret);
 }
 
 bool
@@ -564,36 +530,30 @@ mongoc_bulk_operation_update_many_with_opts (mongoc_bulk_operation_t *bulk,
                                              const bson_t *opts,
                                              bson_error_t *error) /* OUT */
 {
-   bool retval;
-   bson_t opts_dup;
-   bson_iter_t iter;
+   mongoc_bulk_update_many_opts_t update_opts;
+   bool ret;
 
    ENTRY;
 
-   if (opts && bson_iter_init_find (&iter, opts, "multi")) {
-      if (!BSON_ITER_HOLDS_BOOL (&iter) || !bson_iter_bool (&iter)) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_COMMAND_INVALID_ARG,
-                         "%s expects the 'multi' option to be true",
-                         BSON_FUNC);
-         RETURN (false);
-      }
+   BULK_RETURN_IF_PRIOR_ERROR;
 
-      return _mongoc_bulk_operation_update_with_opts (
-         bulk, selector, document, opts, error);
+   if (!_mongoc_bulk_update_many_opts_parse (
+          bulk->client, opts, &update_opts, error)) {
+      _mongoc_bulk_update_many_opts_cleanup (&update_opts);
+      RETURN (false);
    }
 
-   bson_init (&opts_dup);
-   BSON_APPEND_BOOL (&opts_dup, "multi", true);
-   if (opts) {
-      bson_concat (&opts_dup, opts);
-   }
-   retval = _mongoc_bulk_operation_update_with_opts (
-      bulk, selector, document, &opts_dup, error);
-   bson_destroy (&opts_dup);
+   ret = _mongoc_bulk_operation_update_with_opts (bulk,
+                                                  selector,
+                                                  document,
+                                                  &update_opts.update,
+                                                  &update_opts.arrayFilters,
+                                                  &update_opts.extra,
+                                                  true /* multi */,
+                                                  error);
 
-   RETURN (retval);
+   _mongoc_bulk_update_many_opts_cleanup (&update_opts);
+   RETURN (ret);
 }
 
 void
@@ -610,10 +570,11 @@ mongoc_bulk_operation_update (mongoc_bulk_operation_t *bulk,
    BULK_EXIT_IF_PRIOR_ERROR;
 
    bson_init (&opts);
-   BSON_APPEND_BOOL (&opts, "upsert", upsert);
-   BSON_APPEND_BOOL (&opts, "multi", true);
+   if (upsert) {
+      BSON_APPEND_BOOL (&opts, "upsert", upsert);
+   }
 
-   _mongoc_bulk_operation_update_with_opts (
+   mongoc_bulk_operation_update_many_with_opts (
       bulk, selector, document, &opts, error);
 
    bson_destroy (&opts);
@@ -640,10 +601,11 @@ mongoc_bulk_operation_update_one (mongoc_bulk_operation_t *bulk,
 
    bson_init (&opts);
    BSON_APPEND_BOOL (&opts, "upsert", upsert);
-   BSON_APPEND_BOOL (&opts, "multi", false);
 
-   _mongoc_bulk_operation_update_with_opts (
-      bulk, selector, document, &opts, error);
+   if (!mongoc_bulk_operation_update_one_with_opts (
+         bulk, selector, document, &opts, error)) {
+      MONGOC_WARNING ("%s", error->message);
+   }
 
    bson_destroy (&opts);
 
@@ -654,6 +616,77 @@ mongoc_bulk_operation_update_one (mongoc_bulk_operation_t *bulk,
    EXIT;
 }
 
+bool
+mongoc_bulk_operation_replace_one_with_opts (mongoc_bulk_operation_t *bulk,
+                                             const bson_t *selector,
+                                             const bson_t *document,
+                                             const bson_t *opts,
+                                             bson_error_t *error) /* OUT */
+{
+   mongoc_bulk_replace_one_opts_t repl_opts;
+   mongoc_bulk_update_opts_t *update_opts = &repl_opts.update;
+   bool ret = false;
+
+   ENTRY;
+
+   BSON_ASSERT (bulk);
+   BSON_ASSERT (selector);
+   BSON_ASSERT (document);
+
+   BULK_RETURN_IF_PRIOR_ERROR;
+
+   if (!_mongoc_bulk_replace_one_opts_parse (
+          bulk->client, opts, &repl_opts, error)) {
+      GOTO (done);
+   }
+
+   if (!_mongoc_validate_replace (document, update_opts->validate, error)) {
+      GOTO (done);
+   }
+
+   /* allow "multi" in opts, but it must be the correct multi */
+   if (update_opts->multi) {
+      bson_set_error (error,
+                      MONGOC_ERROR_COMMAND,
+                      MONGOC_ERROR_COMMAND_INVALID_ARG,
+                      "Invalid \"multi\": true in opts for"
+                      " mongoc_bulk_operation_replace_one_with_opts."
+                      " The value must be true, or omitted.");
+      GOTO (done);
+   }
+
+   _mongoc_bulk_operation_update_append (
+      bulk, selector, document, update_opts, NULL, &repl_opts.extra);
+   ret = true;
+
+done:
+   _mongoc_bulk_replace_one_opts_cleanup (&repl_opts);
+   RETURN (ret);
+}
+
+void
+mongoc_bulk_operation_replace_one (mongoc_bulk_operation_t *bulk,
+                                   const bson_t *selector,
+                                   const bson_t *document,
+                                   bool upsert)
+{
+   bson_t opts = BSON_INITIALIZER;
+   bson_error_t *error = &bulk->result.error;
+
+   ENTRY;
+
+   BSON_APPEND_BOOL (&opts, "upsert", upsert);
+
+   if (!mongoc_bulk_operation_replace_one_with_opts (
+          bulk, selector, document, &opts, error)) {
+      MONGOC_WARNING ("%s", error->message);
+   }
+
+   bson_destroy (&opts);
+
+   EXIT;
+}
+
 uint32_t
 mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
                                bson_t *reply,                 /* OUT */
@@ -670,39 +703,37 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
 
    BSON_ASSERT (bulk);
 
-   cluster = &bulk->client->cluster;
-
-   if (bulk->executed) {
-      _mongoc_write_result_destroy (&bulk->result);
-   }
-
-   bulk->executed = true;
-
-   if (reply) {
-      bson_init (reply);
-   }
-
    if (!bulk->client) {
       bson_set_error (error,
                       MONGOC_ERROR_COMMAND,
                       MONGOC_ERROR_COMMAND_INVALID_ARG,
                       "mongoc_bulk_operation_execute() requires a client "
                       "and one has not been set.");
-      RETURN (false);
-   } else if (!bulk->database) {
+      GOTO (err);
+   }
+   cluster = &bulk->client->cluster;
+
+   if (bulk->executed) {
+      _mongoc_write_result_destroy (&bulk->result);
+      _mongoc_write_result_init (&bulk->result);
+   }
+
+   bulk->executed = true;
+
+   if (!bulk->database) {
       bson_set_error (error,
                       MONGOC_ERROR_COMMAND,
                       MONGOC_ERROR_COMMAND_INVALID_ARG,
                       "mongoc_bulk_operation_execute() requires a database "
                       "and one has not been set.");
-      RETURN (false);
+      GOTO (err);
    } else if (!bulk->collection) {
       bson_set_error (error,
                       MONGOC_ERROR_COMMAND,
                       MONGOC_ERROR_COMMAND_INVALID_ARG,
                       "mongoc_bulk_operation_execute() requires a collection "
                       "and one has not been set.");
-      RETURN (false);
+      GOTO (err);
    }
 
    /* error stored by functions like mongoc_bulk_operation_insert that
@@ -712,7 +743,7 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
          memcpy (error, &bulk->result.error, sizeof (bson_error_t));
       }
 
-      RETURN (false);
+      GOTO (err);
    }
 
    if (!bulk->commands.len) {
@@ -720,21 +751,28 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
                       MONGOC_ERROR_COMMAND,
                       MONGOC_ERROR_COMMAND_INVALID_ARG,
                       "Cannot do an empty bulk write");
-      RETURN (false);
-   }
-
-   if (bulk->server_id) {
-      server_stream = mongoc_cluster_stream_for_server (
-         cluster, bulk->server_id, true /* reconnect_ok */, error);
-   } else {
-      server_stream = mongoc_cluster_stream_for_writes (cluster, error);
-   }
-
-   if (!server_stream) {
-      RETURN (false);
+      GOTO (err);
    }
 
    for (i = 0; i < bulk->commands.len; i++) {
+      if (bulk->server_id) {
+         server_stream =
+            mongoc_cluster_stream_for_server (cluster,
+                                              bulk->server_id,
+                                              true /* reconnect_ok */,
+                                              bulk->session,
+                                              reply,
+                                              error);
+      } else {
+         server_stream = mongoc_cluster_stream_for_writes (
+            cluster, bulk->session, reply, error);
+      }
+
+      if (!server_stream) {
+         /* stream_for_server and stream_for_writes initialize reply on error */
+         RETURN (false);
+      }
+
       command =
          &_mongoc_array_index (&bulk->commands, mongoc_write_command_t, i);
 
@@ -745,28 +783,35 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
                                      bulk->collection,
                                      bulk->write_concern,
                                      offset,
+                                     bulk->session,
                                      &bulk->result);
 
       bulk->server_id = server_stream->sd->id;
 
       if (bulk->result.failed &&
           (bulk->flags.ordered || bulk->result.must_stop)) {
+         mongoc_server_stream_cleanup (server_stream);
          GOTO (cleanup);
       }
 
       offset += command->n_documents;
+      mongoc_server_stream_cleanup (server_stream);
    }
 
 cleanup:
-   ret = _mongoc_write_result_complete (&bulk->result,
-                                        bulk->client->error_api_version,
-                                        bulk->write_concern,
-                                        MONGOC_ERROR_COMMAND /* err domain */,
-                                        reply,
-                                        error);
-   mongoc_server_stream_cleanup (server_stream);
+   _mongoc_bson_init_if_set (reply);
+   ret = MONGOC_WRITE_RESULT_COMPLETE (&bulk->result,
+                                       bulk->client->error_api_version,
+                                       bulk->write_concern,
+                                       MONGOC_ERROR_COMMAND /* err domain */,
+                                       reply,
+                                       error);
 
    RETURN (ret ? bulk->server_id : 0);
+
+err:
+   _mongoc_bson_init_if_set (reply);
+   RETURN (false);
 }
 
 void
@@ -827,17 +872,39 @@ void
 mongoc_bulk_operation_set_client (mongoc_bulk_operation_t *bulk, void *client)
 {
    BSON_ASSERT (bulk);
+   BSON_ASSERT (client);
+
+   if (bulk->session) {
+      BSON_ASSERT (bulk->session->client == client);
+   }
 
    bulk->client = (mongoc_client_t *) client;
 
    /* if you call set_client, bulk was likely made by mongoc_bulk_operation_new,
-    * not mongoc_collection_create_bulk_operation(), so operation_id is 0. */
+    * not mongoc_collection_create_bulk_operation_with_opts(), so operation_id
+    * is 0. */
    if (!bulk->operation_id) {
       bulk->operation_id = ++bulk->client->cluster.operation_id;
    }
 }
 
 
+void
+mongoc_bulk_operation_set_client_session (
+   mongoc_bulk_operation_t *bulk,
+   struct _mongoc_client_session_t *client_session)
+{
+   BSON_ASSERT (bulk);
+   BSON_ASSERT (client_session);
+
+   if (bulk->client) {
+      BSON_ASSERT (bulk->client == client_session->client);
+   }
+
+   bulk->session = client_session;
+}
+
+
 uint32_t
 mongoc_bulk_operation_get_hint (const mongoc_bulk_operation_t *bulk)
 {
