@@ -19,9 +19,6 @@
 #include "mongoc-collection.h"
 #include "mongoc-collection-private.h"
 #include "mongoc-cursor.h"
-#include "mongoc-cursor-array-private.h"
-#include "mongoc-cursor-cursorid-private.h"
-#include "mongoc-cursor-transform-private.h"
 #include "mongoc-cursor-private.h"
 #include "mongoc-database.h"
 #include "mongoc-database-private.h"
@@ -30,7 +27,7 @@
 #include "mongoc-trace-private.h"
 #include "mongoc-util-private.h"
 #include "mongoc-write-concern-private.h"
-
+#include "mongoc-change-stream-private.h"
 
 #undef MONGOC_LOG_DOMAIN
 #define MONGOC_LOG_DOMAIN "database"
@@ -106,7 +103,9 @@ mongoc_database_destroy (mongoc_database_t *database)
 {
    ENTRY;
 
-   BSON_ASSERT (database);
+   if (!database) {
+      EXIT;
+   }
 
    if (database->read_prefs) {
       mongoc_read_prefs_destroy (database->read_prefs);
@@ -169,24 +168,23 @@ mongoc_database_command (mongoc_database_t *database,
                          const bson_t *fields,
                          const mongoc_read_prefs_t *read_prefs)
 {
+   char ns[MONGOC_NAMESPACE_MAX];
+
    BSON_ASSERT (database);
    BSON_ASSERT (command);
 
+   bson_snprintf (ns, sizeof ns, "%s.$cmd", database->name);
+
    /* Server Selection Spec: "The generic command method has a default read
     * preference of mode 'primary'. The generic command method MUST ignore any
     * default read preference from client, database or collection
     * configuration. The generic command method SHOULD allow an optional read
     * preference argument."
     */
-   return mongoc_client_command (database->client,
-                                 database->name,
-                                 flags,
-                                 skip,
-                                 limit,
-                                 batch_size,
-                                 command,
-                                 fields,
-                                 read_prefs);
+
+   /* flags, skip, limit, batch_size, fields are unused */
+   return _mongoc_cursor_cmd_deprecated_new (
+      database->client, ns, command, read_prefs);
 }
 
 
@@ -206,8 +204,19 @@ mongoc_database_command_simple (mongoc_database_t *database,
     * configuration. The generic command method SHOULD allow an optional read
     * preference argument."
     */
-   return mongoc_client_command_simple (
-      database->client, database->name, command, read_prefs, reply, error);
+
+   return _mongoc_client_command_with_opts (database->client,
+                                            database->name,
+                                            command,
+                                            MONGOC_CMD_READ,
+                                            NULL /* opts */,
+                                            MONGOC_QUERY_NONE,
+                                            read_prefs,
+                                            NULL, /* user prefs */
+                                            NULL /* read concern */,
+                                            NULL /* write concern */,
+                                            reply,
+                                            error);
 }
 
 
@@ -219,18 +228,18 @@ mongoc_database_read_command_with_opts (mongoc_database_t *database,
                                         bson_t *reply,
                                         bson_error_t *error)
 {
-   return _mongoc_client_command_with_opts (
-      database->client,
-      database->name,
-      command,
-      MONGOC_CMD_READ,
-      opts,
-      MONGOC_QUERY_NONE,
-      COALESCE (read_prefs, database->read_prefs),
-      database->read_concern,
-      database->write_concern,
-      reply,
-      error);
+   return _mongoc_client_command_with_opts (database->client,
+                                            database->name,
+                                            command,
+                                            MONGOC_CMD_READ,
+                                            opts,
+                                            MONGOC_QUERY_NONE,
+                                            read_prefs,
+                                            database->read_prefs,
+                                            database->read_concern,
+                                            database->write_concern,
+                                            reply,
+                                            error);
 }
 
 
@@ -247,6 +256,7 @@ mongoc_database_write_command_with_opts (mongoc_database_t *database,
                                             MONGOC_CMD_WRITE,
                                             opts,
                                             MONGOC_QUERY_NONE,
+                                            NULL, /* user prefs */
                                             database->read_prefs,
                                             database->read_concern,
                                             database->write_concern,
@@ -259,23 +269,46 @@ bool
 mongoc_database_read_write_command_with_opts (
    mongoc_database_t *database,
    const bson_t *command,
-   const mongoc_read_prefs_t *read_prefs,
+   const mongoc_read_prefs_t *read_prefs /* IGNORED */,
    const bson_t *opts,
    bson_t *reply,
    bson_error_t *error)
 {
-   return _mongoc_client_command_with_opts (
-      database->client,
-      database->name,
-      command,
-      MONGOC_CMD_RW,
-      opts,
-      MONGOC_QUERY_NONE,
-      COALESCE (read_prefs, database->read_prefs),
-      database->read_concern,
-      database->write_concern,
-      reply,
-      error);
+   return _mongoc_client_command_with_opts (database->client,
+                                            database->name,
+                                            command,
+                                            MONGOC_CMD_RW,
+                                            opts,
+                                            MONGOC_QUERY_NONE,
+                                            read_prefs,
+                                            database->read_prefs,
+                                            database->read_concern,
+                                            database->write_concern,
+                                            reply,
+                                            error);
+}
+
+
+bool
+mongoc_database_command_with_opts (mongoc_database_t *database,
+                                   const bson_t *command,
+                                   const mongoc_read_prefs_t *read_prefs,
+                                   const bson_t *opts,
+                                   bson_t *reply,
+                                   bson_error_t *error)
+{
+   return _mongoc_client_command_with_opts (database->client,
+                                            database->name,
+                                            command,
+                                            MONGOC_CMD_RAW,
+                                            opts,
+                                            MONGOC_QUERY_NONE,
+                                            read_prefs,
+                                            NULL, /* default prefs */
+                                            database->read_concern,
+                                            database->write_concern,
+                                            reply,
+                                            error);
 }
 
 
@@ -324,6 +357,7 @@ mongoc_database_drop_with_opts (mongoc_database_t *database,
                                            MONGOC_CMD_WRITE,
                                            opts,
                                            MONGOC_QUERY_NONE,
+                                           NULL, /* user prefs */
                                            database->read_prefs,
                                            database->read_concern,
                                            database->write_concern,
@@ -335,114 +369,11 @@ mongoc_database_drop_with_opts (mongoc_database_t *database,
 }
 
 
-/*
- *--------------------------------------------------------------------------
- *
- * mongoc_database_add_user_legacy --
- *
- *       A helper to add a user or update their password on @database.
- *       This uses the legacy protocol by inserting into system.users.
- *
- * Returns:
- *       true if successful; otherwise false and @error is set.
- *
- * Side effects:
- *       @error may be set.
- *
- *--------------------------------------------------------------------------
- */
-
-static bool
-mongoc_database_add_user_legacy (mongoc_database_t *database,
-                                 const char *username,
-                                 const char *password,
-                                 bson_error_t *error)
-{
-   mongoc_collection_t *collection;
-   mongoc_cursor_t *cursor = NULL;
-   const bson_t *doc;
-   bool ret = false;
-   bson_t query;
-   bson_t opts;
-   bson_t user;
-   char *input;
-   char *pwd = NULL;
-
-   ENTRY;
-
-   BSON_ASSERT (database);
-   BSON_ASSERT (username);
-   BSON_ASSERT (password);
-
-   /*
-    * Users are stored in the <dbname>.system.users virtual collection.
-    */
-   collection = mongoc_client_get_collection (
-      database->client, database->name, "system.users");
-   BSON_ASSERT (collection);
-
-   /*
-    * Hash the users password.
-    */
-   input = bson_strdup_printf ("%s:mongo:%s", username, password);
-   pwd = _mongoc_hex_md5 (input);
-   bson_free (input);
-
-   /*
-    * Check to see if the user exists. If so, we will update the
-    * password instead of inserting a new user.
-    */
-   bson_init (&query);
-   bson_append_utf8 (&query, "user", 4, username, -1);
-
-   bson_init (&opts);
-   bson_append_int64 (&opts, "limit", 5, 1);
-   bson_append_bool (&opts, "singleBatch", 11, true);
-
-   cursor = mongoc_collection_find_with_opts (collection, &query, &opts, NULL);
-   if (!mongoc_cursor_next (cursor, &doc)) {
-      if (mongoc_cursor_error (cursor, error)) {
-         GOTO (failure);
-      }
-      bson_init (&user);
-      bson_append_utf8 (&user, "user", 4, username, -1);
-      bson_append_bool (&user, "readOnly", 8, false);
-      bson_append_utf8 (&user, "pwd", 3, pwd, -1);
-   } else {
-      bson_init (&user);
-      bson_copy_to_excluding_noinit (doc, &user, "pwd", (char *) NULL);
-      bson_append_utf8 (&user, "pwd", 3, pwd, -1);
-   }
-
-   if (!mongoc_collection_insert (collection, MONGOC_INSERT_NONE, &user, NULL, error)) {
-      GOTO (failure_with_user);
-   }
-
-   ret = true;
-
-failure_with_user:
-   bson_destroy (&user);
-
-failure:
-   if (cursor) {
-      mongoc_cursor_destroy (cursor);
-   }
-   mongoc_collection_destroy (collection);
-   bson_destroy (&query);
-   bson_destroy (&opts);
-   bson_free (pwd);
-
-   RETURN (ret);
-}
-
-
 bool
 mongoc_database_remove_user (mongoc_database_t *database,
                              const char *username,
                              bson_error_t *error)
 {
-   mongoc_collection_t *col;
-   bson_error_t lerror;
    bson_t cmd;
    bool ret;
 
@@ -453,26 +384,9 @@ mongoc_database_remove_user (mongoc_database_t *database,
 
    bson_init (&cmd);
    BSON_APPEND_UTF8 (&cmd, "dropUser", username);
-   ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, &lerror);
+   ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, error);
    bson_destroy (&cmd);
 
-   if (!ret && (lerror.code == MONGOC_ERROR_QUERY_COMMAND_NOT_FOUND)) {
-      bson_init (&cmd);
-      BSON_APPEND_UTF8 (&cmd, "user", username);
-
-      col = mongoc_client_get_collection (
-         database->client, database->name, "system.users");
-      BSON_ASSERT (col);
-
-      ret = mongoc_collection_remove (
-         col, MONGOC_REMOVE_SINGLE_REMOVE, &cmd, NULL, error);
-
-      bson_destroy (&cmd);
-      mongoc_collection_destroy (col);
-   } else if (error) {
-      memcpy (error, &lerror, sizeof *error);
-   }
-
    RETURN (ret);
 }
 
@@ -481,8 +395,6 @@ bool
 mongoc_database_remove_all_users (mongoc_database_t *database,
                                   bson_error_t *error)
 {
-   mongoc_collection_t *col;
-   bson_error_t lerror;
    bson_t cmd;
    bool ret;
 
@@ -492,25 +404,9 @@ mongoc_database_remove_all_users (mongoc_database_t *database,
 
    bson_init (&cmd);
    BSON_APPEND_INT32 (&cmd, "dropAllUsersFromDatabase", 1);
-   ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, &lerror);
+   ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, error);
    bson_destroy (&cmd);
 
-   if (!ret && (lerror.code == MONGOC_ERROR_QUERY_COMMAND_NOT_FOUND)) {
-      bson_init (&cmd);
-
-      col = mongoc_client_get_collection (
-         database->client, database->name, "system.users");
-      BSON_ASSERT (col);
-
-      ret =
-         mongoc_collection_remove (col, MONGOC_REMOVE_NONE, &cmd, NULL, error);
-
-      bson_destroy (&cmd);
-      mongoc_collection_destroy (col);
-   } else if (error) {
-      memcpy (error, &lerror, sizeof *error);
-   }
-
    RETURN (ret);
 }
 
@@ -537,11 +433,8 @@ mongoc_database_add_user (mongoc_database_t *database,
                           const bson_t *custom_data,
                           bson_error_t *error)
 {
-   bson_error_t lerror;
    bson_t cmd;
    bson_t ar;
-   char *input;
-   char *hashed_password;
    bool ret = false;
 
    ENTRY;
@@ -549,48 +442,23 @@ mongoc_database_add_user (mongoc_database_t *database,
    BSON_ASSERT (database);
    BSON_ASSERT (username);
 
-   /*
-    * CDRIVER-232:
-    *
-    * Perform a (slow and tedious) round trip to mongod to determine if
-    * we can safely call createUser. Otherwise, we will fallback and
-    * perform legacy insertion into users collection.
-    */
    bson_init (&cmd);
-   BSON_APPEND_UTF8 (&cmd, "usersInfo", username);
-   ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, &lerror);
-   bson_destroy (&cmd);
+   BSON_APPEND_UTF8 (&cmd, "createUser", username);
+   BSON_APPEND_UTF8 (&cmd, "pwd", password);
 
-   if (!ret && (lerror.code == MONGOC_ERROR_QUERY_COMMAND_NOT_FOUND)) {
-      ret =
-         mongoc_database_add_user_legacy (database, username, password, error);
-   } else if (ret || (lerror.code == 13)) {
-      /* usersInfo succeeded or failed with auth err, we're on modern mongod */
-      input = bson_strdup_printf ("%s:mongo:%s", username, password);
-      hashed_password = _mongoc_hex_md5 (input);
-      bson_free (input);
-
-      bson_init (&cmd);
-      BSON_APPEND_UTF8 (&cmd, "createUser", username);
-      BSON_APPEND_UTF8 (&cmd, "pwd", hashed_password);
-      BSON_APPEND_BOOL (&cmd, "digestPassword", false);
-      if (custom_data) {
-         BSON_APPEND_DOCUMENT (&cmd, "customData", custom_data);
-      }
-      if (roles) {
-         BSON_APPEND_ARRAY (&cmd, "roles", roles);
-      } else {
-         bson_append_array_begin (&cmd, "roles", 5, &ar);
-         bson_append_array_end (&cmd, &ar);
-      }
-
-      ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, error);
-
-      bson_free (hashed_password);
-      bson_destroy (&cmd);
-   } else if (error) {
-      memcpy (error, &lerror, sizeof *error);
+   if (custom_data) {
+      BSON_APPEND_DOCUMENT (&cmd, "customData", custom_data);
    }
+   if (roles) {
+      BSON_APPEND_ARRAY (&cmd, "roles", roles);
+   } else {
+      bson_append_array_begin (&cmd, "roles", 5, &ar);
+      bson_append_array_end (&cmd, &ar);
+   }
+
+   ret = mongoc_database_command_simple (database, &cmd, NULL, NULL, error);
+
+   bson_destroy (&cmd);
 
    RETURN (ret);
 }
@@ -794,7 +662,8 @@ mongoc_database_has_collection (mongoc_database_t *database,
    bson_iter_t col_iter;
    bool ret = false;
    const char *cur_name;
-   bson_t filter = BSON_INITIALIZER;
+   bson_t opts = BSON_INITIALIZER;
+   bson_t filter;
    mongoc_cursor_t *cursor;
    const bson_t *doc;
 
@@ -807,18 +676,11 @@ mongoc_database_has_collection (mongoc_database_t *database,
       memset (error, 0, sizeof *error);
    }
 
+   BSON_APPEND_DOCUMENT_BEGIN (&opts, "filter", &filter);
    BSON_APPEND_UTF8 (&filter, "name", name);
+   bson_append_document_end (&opts, &filter);
 
-   cursor = mongoc_database_find_collections (database, &filter, error);
-
-   if (!cursor) {
-      return ret;
-   }
-
-   if (error && ((error->domain != 0) || (error->code != 0))) {
-      GOTO (cleanup);
-   }
-
+   cursor = mongoc_database_find_collections_with_opts (database, &opts);
    while (mongoc_cursor_next (cursor, &doc)) {
       if (bson_iter_init (&col_iter, doc) &&
           bson_iter_find (&col_iter, "name") &&
@@ -831,189 +693,90 @@ mongoc_database_has_collection (mongoc_database_t *database,
       }
    }
 
+   (void) mongoc_cursor_error (cursor, error);
+
 cleanup:
    mongoc_cursor_destroy (cursor);
+   bson_destroy (&opts);
 
    RETURN (ret);
 }
 
-typedef struct {
-   const char *dbname;
-   size_t dbname_len;
-   const char *name;
-} mongoc_database_find_collections_legacy_ctx_t;
-
-static mongoc_cursor_transform_mode_t
-_mongoc_database_find_collections_legacy_filter (const bson_t *bson, void *ctx_)
-{
-   bson_iter_t iter;
-   mongoc_database_find_collections_legacy_ctx_t *ctx;
-
-   ctx = (mongoc_database_find_collections_legacy_ctx_t *) ctx_;
-
-   if (bson_iter_init_find (&iter, bson, "name") &&
-       BSON_ITER_HOLDS_UTF8 (&iter) &&
-       (ctx->name = bson_iter_utf8 (&iter, NULL)) && !strchr (ctx->name, '$') &&
-       (0 == strncmp (ctx->name, ctx->dbname, ctx->dbname_len))) {
-      return MONGO_CURSOR_TRANSFORM_MUTATE;
-   } else {
-      return MONGO_CURSOR_TRANSFORM_DROP;
-   }
-}
-
-static void
-_mongoc_database_find_collections_legacy_mutate (const bson_t *bson,
-                                                 bson_t *out,
-                                                 void *ctx_)
-{
-   mongoc_database_find_collections_legacy_ctx_t *ctx;
-
-   ctx = (mongoc_database_find_collections_legacy_ctx_t *) ctx_;
-
-   bson_copy_to_excluding_noinit (bson, out, "name", NULL);
-   BSON_APPEND_UTF8 (
-      out, "name", ctx->name + (ctx->dbname_len + 1)); /* +1 for the '.' */
-}
-
-/* Uses old way of querying system.namespaces. */
-mongoc_cursor_t *
-_mongoc_database_find_collections_legacy (mongoc_database_t *database,
-                                          const bson_t *filter,
-                                          bson_error_t *error)
-{
-   mongoc_collection_t *col;
-   mongoc_cursor_t *cursor = NULL;
-   mongoc_read_prefs_t *read_prefs;
-   uint32_t dbname_len;
-   bson_t legacy_filter;
-   bson_iter_t iter;
-   const char *col_filter;
-   bson_t q = BSON_INITIALIZER;
-   mongoc_database_find_collections_legacy_ctx_t *ctx;
-
-   BSON_ASSERT (database);
-
-   col = mongoc_client_get_collection (
-      database->client, database->name, "system.namespaces");
-
-   BSON_ASSERT (col);
-
-   dbname_len = (uint32_t) strlen (database->name);
-
-   ctx = (mongoc_database_find_collections_legacy_ctx_t *) bson_malloc (
-      sizeof (*ctx));
-
-   ctx->dbname = database->name;
-   ctx->dbname_len = dbname_len;
-
-   /* Filtering on name needs to be handled differently for old servers. */
-   if (filter && bson_iter_init_find (&iter, filter, "name")) {
-      bson_string_t *buf;
-      /* on legacy servers, this must be a string (i.e. not a regex) */
-      if (!BSON_ITER_HOLDS_UTF8 (&iter)) {
-         bson_set_error (
-            error,
-            MONGOC_ERROR_NAMESPACE,
-            MONGOC_ERROR_NAMESPACE_INVALID_FILTER_TYPE,
-            "On legacy servers, a filter on name can only be a string.");
-         bson_free (ctx);
-         goto cleanup_filter;
-      }
-      BSON_ASSERT (BSON_ITER_HOLDS_UTF8 (&iter));
-      col_filter = bson_iter_utf8 (&iter, NULL);
-      bson_init (&legacy_filter);
-      bson_copy_to_excluding_noinit (filter, &legacy_filter, "name", NULL);
-      /* We must db-qualify filters on name. */
-      buf = bson_string_new (database->name);
-      bson_string_append_c (buf, '.');
-      bson_string_append (buf, col_filter);
-      BSON_APPEND_UTF8 (&legacy_filter, "name", buf->str);
-      bson_string_free (buf, true);
-      filter = &legacy_filter;
-   }
-
-   /* Enumerate Collections Spec: "run listCollections on the primary node in
-    * replicaset mode" */
-   read_prefs = mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
-
-   cursor = mongoc_collection_find_with_opts (
-      col, filter ? filter : &q, NULL, read_prefs);
-
-   _mongoc_cursor_transform_init (
-      cursor,
-      _mongoc_database_find_collections_legacy_filter,
-      _mongoc_database_find_collections_legacy_mutate,
-      &bson_free,
-      ctx);
-
-   mongoc_read_prefs_destroy (read_prefs);
-
-cleanup_filter:
-   mongoc_collection_destroy (col);
-
-   return cursor;
-}
 
 mongoc_cursor_t *
 mongoc_database_find_collections (mongoc_database_t *database,
                                   const bson_t *filter,
                                   bson_error_t *error)
+{
+   bson_t opts = BSON_INITIALIZER;
+   mongoc_cursor_t *cursor;
+
+   BSON_ASSERT (database);
+
+   if (filter) {
+      if (!BSON_APPEND_DOCUMENT (&opts, "filter", filter)) {
+         bson_set_error (error,
+                         MONGOC_ERROR_BSON,
+                         MONGOC_ERROR_BSON_INVALID,
+                         "Invalid 'filter' parameter.");
+         bson_destroy (&opts);
+         return NULL;
+      }
+   }
+
+   cursor = mongoc_database_find_collections_with_opts (database, &opts);
+
+   bson_destroy (&opts);
+
+   /* this deprecated API returns NULL on error */
+   if (mongoc_cursor_error (cursor, error)) {
+      mongoc_cursor_destroy (cursor);
+      return NULL;
+   }
+
+   return cursor;
+}
+
+
+mongoc_cursor_t *
+mongoc_database_find_collections_with_opts (mongoc_database_t *database,
+                                            const bson_t *opts)
 {
    mongoc_cursor_t *cursor;
    bson_t cmd = BSON_INITIALIZER;
-   bson_t child;
-   bson_error_t lerror;
 
    BSON_ASSERT (database);
 
    BSON_APPEND_INT32 (&cmd, "listCollections", 1);
 
-   if (filter) {
-      BSON_APPEND_DOCUMENT (&cmd, "filter", filter);
-      BSON_APPEND_DOCUMENT_BEGIN (&cmd, "cursor", &child);
-      bson_append_document_end (&cmd, &child);
-   }
-
    /* Enumerate Collections Spec: "run listCollections on the primary node in
     * replicaset mode" */
-   cursor = _mongoc_cursor_new_with_opts (database->client,
-                                          database->name,
-                                          true /* is_command */,
-                                          NULL,
-                                          NULL,
-                                          NULL,
-                                          NULL);
-
-   _mongoc_cursor_cursorid_init (cursor, &cmd);
-
-   if (_mongoc_cursor_cursorid_prime (cursor)) {
-      /* intentionally empty */
-   } else {
-      if (mongoc_cursor_error (cursor, &lerror)) {
-         if (lerror.code == MONGOC_ERROR_QUERY_COMMAND_NOT_FOUND) {
-            /* We are talking to a server that doesn' support listCollections.
-             */
-            /* clear out the error. */
-            memset (&lerror, 0, sizeof lerror);
-            /* try again with using system.namespaces */
-            mongoc_cursor_destroy (cursor);
-            cursor = _mongoc_database_find_collections_legacy (
-               database, filter, error);
-         } else if (error) {
-            memcpy (error, &lerror, sizeof *error);
-         }
-      }
+   cursor = _mongoc_cursor_cmd_new (
+      database->client, database->name, &cmd, opts, NULL, NULL, NULL);
+   if (cursor->error.domain == 0) {
+      _mongoc_cursor_prime (cursor);
    }
-
    bson_destroy (&cmd);
 
    return cursor;
 }
 
+
 char **
 mongoc_database_get_collection_names (mongoc_database_t *database,
                                       bson_error_t *error)
 {
+   return mongoc_database_get_collection_names_with_opts (
+      database, NULL, error);
+}
+
+
+char **
+mongoc_database_get_collection_names_with_opts (mongoc_database_t *database,
+                                                const bson_t *opts,
+                                                bson_error_t *error)
+{
+   bson_t opts_copy;
    bson_iter_t col;
    const char *name;
    char *namecopy;
@@ -1024,12 +787,20 @@ mongoc_database_get_collection_names (mongoc_database_t *database,
 
    BSON_ASSERT (database);
 
-   cursor = mongoc_database_find_collections (database, NULL, error);
-
-   if (!cursor) {
-      return NULL;
+   if (opts) {
+      bson_copy_to (opts, &opts_copy);
+   } else {
+      bson_init (&opts_copy);
    }
 
+   /* nameOnly option is faster in MongoDB 4+, ignored by older versions,
+    * see Enumerating Collections Spec */
+   if (!bson_has_field (&opts_copy, "nameOnly")) {
+      bson_append_bool (&opts_copy, "nameOnly", 8, true);
+   }
+
+   cursor = mongoc_database_find_collections_with_opts (database, &opts_copy);
+
    _mongoc_array_init (&strv_buf, sizeof (char *));
 
    while (mongoc_cursor_next (cursor, &doc)) {
@@ -1053,6 +824,7 @@ mongoc_database_get_collection_names (mongoc_database_t *database,
    }
 
    mongoc_cursor_destroy (cursor);
+   bson_destroy (&opts_copy);
 
    return ret;
 }
@@ -1094,7 +866,7 @@ mongoc_database_create_collection (mongoc_database_t *database,
       }
 
       if (bson_iter_init_find (&iter, opts, "size")) {
-         if (!BSON_ITER_HOLDS_INT32 (&iter) && !BSON_ITER_HOLDS_INT64 (&iter)) {
+         if (!BSON_ITER_HOLDS_INT (&iter)) {
             bson_set_error (error,
                             MONGOC_ERROR_COMMAND,
                             MONGOC_ERROR_COMMAND_INVALID_ARG,
@@ -1112,7 +884,7 @@ mongoc_database_create_collection (mongoc_database_t *database,
       }
 
       if (bson_iter_init_find (&iter, opts, "max")) {
-         if (!BSON_ITER_HOLDS_INT32 (&iter) && !BSON_ITER_HOLDS_INT64 (&iter)) {
+         if (!BSON_ITER_HOLDS_INT (&iter)) {
             bson_set_error (error,
                             MONGOC_ERROR_COMMAND,
                             MONGOC_ERROR_COMMAND_INVALID_ARG,
@@ -1181,6 +953,7 @@ mongoc_database_create_collection (mongoc_database_t *database,
                                          MONGOC_CMD_WRITE,
                                          opts,
                                          MONGOC_QUERY_NONE,
+                                         NULL, /* user prefs */
                                          database->read_prefs,
                                          database->read_concern,
                                          database->write_concern,
@@ -1223,3 +996,12 @@ mongoc_database_get_name (mongoc_database_t *database)
 
    return database->name;
 }
+
+
+mongoc_change_stream_t *
+mongoc_database_watch (const mongoc_database_t *db,
+                       const bson_t *pipeline,
+                       const bson_t *opts)
+{
+   return _mongoc_change_stream_new_from_database (db, pipeline, opts);
+}
