@@ -48,6 +48,9 @@ _mongoc_openssl_thread_startup (void);
 static void
 _mongoc_openssl_thread_cleanup (void);
 #endif
+#ifndef MONGOC_HAVE_ASN1_STRING_GET0_DATA
+#define ASN1_STRING_get0_data ASN1_STRING_data
+#endif
 
 /**
  * _mongoc_openssl_init:
@@ -70,11 +73,6 @@ _mongoc_openssl_init (void)
    _mongoc_openssl_thread_startup ();
 #endif
 
-   /*
-    * Ensure we also load the ciphers now from the primary thread
-    * or we can run into some weirdness on 64-bit Solaris 10 on
-    * SPARC with openssl 0.9.7.
-    */
    ctx = SSL_CTX_new (SSLv23_method ());
    if (!ctx) {
       MONGOC_ERROR ("Failed to initialize OpenSSL.");
@@ -132,9 +130,7 @@ _mongoc_openssl_import_cert_store (LPWSTR store_name,
                      (LPTSTR) &msg,
                      0,
                      NULL);
-      MONGOC_ERROR ("Can't open CA store: 0x%.8X: '%s'",
-                    GetLastError (),
-                    msg);
+      MONGOC_ERROR ("Can't open CA store: 0x%.8X: '%s'", (unsigned int) GetLastError (), msg);
       LocalFree (msg);
       return false;
    }
@@ -170,13 +166,14 @@ _mongoc_openssl_import_cert_stores (SSL_CTX *context)
       return false;
    }
 
-   retval = _mongoc_openssl_import_cert_store (
-      L"root", CERT_SYSTEM_STORE_CURRENT_USER|CERT_STORE_READONLY_FLAG, store);
-   if (retval) {
-      return retval;
-   }
-   return _mongoc_openssl_import_cert_store (
-      L"CA", CERT_SYSTEM_STORE_CURRENT_USER|CERT_STORE_READONLY_FLAG, store);
+   retval = _mongoc_openssl_import_cert_store (L"root",
+                                               CERT_SYSTEM_STORE_CURRENT_USER |
+                                                  CERT_STORE_READONLY_FLAG,
+                                               store);
+   retval &= _mongoc_openssl_import_cert_store (
+      L"CA", CERT_SYSTEM_STORE_CURRENT_USER | CERT_STORE_READONLY_FLAG, store);
+
+   return retval;
 }
 #endif
 
@@ -255,7 +252,6 @@ _mongoc_openssl_check_cert (SSL *ssl,
    X509_NAME *subject_name;
    X509_NAME_ENTRY *entry;
    ASN1_STRING *entry_data;
-   char *check;
    int length;
    int idx;
    int r = 0;
@@ -313,7 +309,9 @@ _mongoc_openssl_check_cert (SSL *ssl,
             /* skip entries that can't apply, I.e. IP entries if we've got a
              * DNS host */
             if (name->type == target) {
-               check = (char *) ASN1_STRING_data (name->d.ia5);
+               const char *check;
+
+               check = (const char *) ASN1_STRING_get0_data (name->d.ia5);
                length = ASN1_STRING_length (name->d.ia5);
 
                switch (target) {
@@ -362,6 +360,8 @@ _mongoc_openssl_check_cert (SSL *ssl,
                entry_data = X509_NAME_ENTRY_get_data (entry);
 
                if (entry_data) {
+                  char *check;
+
                   /* TODO: I've heard tell that old versions of SSL crap out
                    * when calling ASN1_STRING_to_UTF8 on already utf8 data.
                    * Check up on that */
@@ -658,6 +658,9 @@ _mongoc_openssl_thread_cleanup (void)
    if (CRYPTO_get_locking_callback () ==
        _mongoc_openssl_thread_locking_callback) {
       CRYPTO_set_locking_callback (NULL);
+   }
+
+   if (CRYPTO_get_id_callback () == _mongoc_openssl_thread_id_callback) {
       CRYPTO_set_id_callback (NULL);
    }
 
