@@ -33,6 +33,11 @@
 #undef MONGOC_LOG_DOMAIN
 #define MONGOC_LOG_DOMAIN "stream-secure-channel"
 
+/* mingw doesn't define this */
+#ifndef SECBUFFER_ALERT
+#define SECBUFFER_ALERT 17
+#endif
+
 
 PCCERT_CONTEXT
 mongoc_secure_channel_setup_certificate_from_file (const char *filename)
@@ -109,7 +114,7 @@ mongoc_secure_channel_setup_certificate_from_file (const char *filename)
       NULL, /* phCertStore, OUT, HCERTSTORE.., unused, for now */
       NULL, /* phMsg, OUT, HCRYPTMSG, only for PKC7, unused */
       (const void **) &cert /* ppvContext, OUT, the Certificate Context */
-      );
+   );
 
    if (!cert) {
       MONGOC_ERROR ("Failed to extract public key from '%s'. Error 0x%.8X",
@@ -160,8 +165,19 @@ mongoc_secure_channel_setup_certificate_from_file (const char *filename)
       NULL,                                    /* pvStructInfo */
       &blob_private_len);                      /* pcbStructInfo */
    if (!success) {
-      MONGOC_ERROR ("Failed to parse private key. Error 0x%.8X",
-                    GetLastError ());
+      LPTSTR msg = NULL;
+      FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER |
+                        FORMAT_MESSAGE_FROM_SYSTEM |
+                        FORMAT_MESSAGE_ARGUMENT_ARRAY,
+                     NULL,
+                     GetLastError (),
+                     LANG_NEUTRAL,
+                     (LPTSTR) &msg,
+                     0,
+                     NULL);
+      MONGOC_ERROR (
+         "Failed to parse private key. %s (0x%.8X)", msg, GetLastError ());
+      LocalFree (msg);
       goto fail;
    }
 
@@ -215,7 +231,7 @@ mongoc_secure_channel_setup_certificate_from_file (const char *filename)
       0,                            /* dwFlags */
       (const void *) provider);     /* pvData */
    if (success) {
-      TRACE ("Successfully loaded client certificate");
+      TRACE ("%s", "Successfully loaded client certificate");
       return cert;
    }
 
@@ -373,7 +389,7 @@ mongoc_secure_channel_setup_ca (
 
    if (CertAddCertificateContextToStore (
           cert_store, cert, CERT_STORE_ADD_USE_EXISTING, NULL)) {
-      TRACE ("Added the certificate !");
+      TRACE ("%s", "Added the certificate !");
       CertCloseStore (cert_store, 0);
       return true;
    }
@@ -415,7 +431,7 @@ mongoc_secure_channel_setup_crl (
       NULL, /* phCertStore, OUT, HCERTSTORE.., unused, for now */
       NULL, /* phMsg, OUT, HCRYPTMSG, only for PKC7, unused */
       (const void **) &cert /* ppvContext, OUT, the Certificate Context */
-      );
+   );
    bson_free (str);
 
    if (!cert) {
@@ -439,7 +455,7 @@ mongoc_secure_channel_setup_crl (
 
    if (CertAddCertificateContextToStore (
           cert_store, cert, CERT_STORE_ADD_USE_EXISTING, NULL)) {
-      TRACE ("Added the certificate !");
+      TRACE ("%s", "Added the certificate !");
       CertFreeCertificateContext (cert);
       CertCloseStore (cert_store, 0);
       return true;
@@ -460,7 +476,7 @@ mongoc_secure_channel_read (mongoc_stream_tls_t *tls,
    ssize_t length;
 
    errno = 0;
-   TRACE ("Wanting to read: %d", data_length);
+   TRACE ("Wanting to read: %d, timeout is %d", data_length, tls->timeout_msec);
    /* 4th argument is minimum bytes, while the data_length is the
     * size of the buffer. We are totally fine with just one TLS record (few
     *bytes)
@@ -494,6 +510,13 @@ mongoc_secure_channel_write (mongoc_stream_tls_t *tls,
    return length;
 }
 
+void
+mongoc_secure_channel_realloc_buf (size_t *size, uint8_t **buf, size_t new_size)
+{
+   *size = bson_next_power_of_two (new_size);
+   *buf = bson_realloc (*buf, *size);
+}
+
 /**
  * The follow functions comes from one of my favorite project, cURL!
  * Thank you so much for having gone through the Secure Channel pain for me.
@@ -588,7 +611,7 @@ mongoc_secure_channel_handshake_step_1 (mongoc_stream_tls_t *tls,
       &outbuf_desc,                       /* pOutput OUT param */
       &secure_channel->ret_flags,         /* pfContextAttr OUT param */
       &secure_channel->ctxt->time_stamp   /* ptsExpiry OUT param */
-      );
+   );
 
    if (sspi_status != SEC_I_CONTINUE_NEEDED) {
       MONGOC_ERROR ("initial InitializeSecurityContext failed: %d",
@@ -631,9 +654,7 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
    mongoc_stream_tls_secure_channel_t *secure_channel =
       (mongoc_stream_tls_secure_channel_t *) tls->ctx;
    SECURITY_STATUS sspi_status = SEC_E_OK;
-   unsigned char *reallocated_buffer;
    ssize_t nread = -1, written = -1;
-   size_t reallocated_length;
    SecBufferDesc outbuf_desc;
    SecBufferDesc inbuf_desc;
    SecBuffer outbuf[3];
@@ -644,39 +665,17 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
    doread = (secure_channel->connecting_state != ssl_connect_2_writing) ? true
                                                                         : false;
 
-   TRACE ("SSL/TLS connection with endpoint (step 2/3)");
+   TRACE ("%s", "SSL/TLS connection with endpoint (step 2/3)");
 
    if (!secure_channel->cred || !secure_channel->ctxt) {
       return false;
    }
 
-   /* buffer to store previously received and decrypted data */
-   if (secure_channel->decdata_buffer == NULL) {
-      secure_channel->decdata_offset = 0;
-      secure_channel->decdata_length = MONGOC_SCHANNEL_BUFFER_INIT_SIZE;
-      secure_channel->decdata_buffer =
-         bson_malloc0 (secure_channel->decdata_length);
-   }
-
-   /* buffer to store previously received and encrypted data */
-   if (secure_channel->encdata_buffer == NULL) {
-      secure_channel->encdata_offset = 0;
-      secure_channel->encdata_length = MONGOC_SCHANNEL_BUFFER_INIT_SIZE;
-      secure_channel->encdata_buffer =
-         bson_malloc0 (secure_channel->encdata_length);
-   }
-
-   /* if we need a bigger buffer to read a full message, increase buffer now */
-   if (secure_channel->encdata_length - secure_channel->encdata_offset <
-       MONGOC_SCHANNEL_BUFFER_FREE_SIZE) {
-      /* increase internal encrypted data buffer */
-      reallocated_length =
-         secure_channel->encdata_offset + MONGOC_SCHANNEL_BUFFER_FREE_SIZE;
-      reallocated_buffer =
-         bson_realloc (secure_channel->encdata_buffer, reallocated_length);
-
-      secure_channel->encdata_buffer = reallocated_buffer;
-      secure_channel->encdata_length = reallocated_length;
+   /* grow the buffer if necessary */
+   if (secure_channel->encdata_length == secure_channel->encdata_offset) {
+      mongoc_secure_channel_realloc_buf (&secure_channel->encdata_length,
+                                         &secure_channel->encdata_buffer,
+                                         secure_channel->encdata_length + 1);
    }
 
    for (;;) {
@@ -694,7 +693,7 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
                   secure_channel->connecting_state = ssl_connect_2_reading;
                }
 
-               TRACE ("failed to receive handshake, need more data");
+               TRACE ("%s", "failed to receive handshake, need more data");
                return true;
             }
 
@@ -763,7 +762,7 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
       /* check if the handshake was incomplete */
       if (sspi_status == SEC_E_INCOMPLETE_MESSAGE) {
          secure_channel->connecting_state = ssl_connect_2_reading;
-         TRACE ("received incomplete message, need more data");
+         TRACE ("%s", "received incomplete message, need more data");
          return true;
       }
 
@@ -774,7 +773,7 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
           !(secure_channel->req_flags & ISC_REQ_USE_SUPPLIED_CREDS)) {
          secure_channel->req_flags |= ISC_REQ_USE_SUPPLIED_CREDS;
          secure_channel->connecting_state = ssl_connect_2_writing;
-         MONGOC_WARNING ("a client certificate has been requested");
+         TRACE ("%s", "A client certificate has been requested");
          return true;
       }
 
@@ -845,6 +844,7 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
 
          default: {
             LPTSTR msg = NULL;
+
             FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER |
                               FORMAT_MESSAGE_FROM_SYSTEM |
                               FORMAT_MESSAGE_ARGUMENT_ARRAY,
@@ -910,7 +910,7 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
    /* check if the handshake is complete */
    if (sspi_status == SEC_E_OK) {
       secure_channel->connecting_state = ssl_connect_3;
-      TRACE ("SSL/TLS handshake complete\n");
+      TRACE ("%s", "SSL/TLS handshake complete");
    }
 
    return true;
@@ -925,7 +925,7 @@ mongoc_secure_channel_handshake_step_3 (mongoc_stream_tls_t *tls,
 
    BSON_ASSERT (ssl_connect_3 == secure_channel->connecting_state);
 
-   TRACE ("SSL/TLS connection with %s (step 3/3)\n", hostname);
+   TRACE ("SSL/TLS connection with %s (step 3/3)", hostname);
 
    if (!secure_channel->cred) {
       return false;
