@@ -1,3 +1,6 @@
+#define R_NO_REMAP
+#define STRICT_R_HEADERS
+
 #include <Rinternals.h>
 #include <bson.h>
 #include <mongoc.h>
