@@ -8,11 +8,11 @@ void bson_reader_finalize(void *handle){
 ssize_t bson_reader_feed(void *handle, void *buf, size_t count){
   int err;
   SEXP *con = (SEXP*) handle;
-  SEXP x = PROTECT(LCONS(ScalarInteger(count), R_NilValue));
-  SEXP y = PROTECT(LCONS(mkString("raw"), x));
-  SEXP z = PROTECT(LCONS(*con, y));
-  SEXP readbin = PROTECT(install("readBin"));
-  SEXP call = PROTECT(LCONS(readbin, z));
+  SEXP x = PROTECT(Rf_lcons(Rf_ScalarInteger(count), R_NilValue));
+  SEXP y = PROTECT(Rf_lcons(Rf_mkString("raw"), x));
+  SEXP z = PROTECT(Rf_lcons(*con, y));
+  SEXP readbin = PROTECT(Rf_install("readBin"));
+  SEXP call = PROTECT(Rf_lcons(readbin, z));
   SEXP res = PROTECT(R_tryEval(call, R_GlobalEnv, &err));
 
   // check if readBin succeeded
@@ -20,13 +20,13 @@ ssize_t bson_reader_feed(void *handle, void *buf, size_t count){
     Rf_error("Mongo reader failed to read data from connection. (%d)", err);
 
   // Copy data into buf
-  memcpy(buf, RAW(res), length(res));
+  memcpy(buf, RAW(res), Rf_length(res));
   UNPROTECT(6);
   return Rf_length(res);
 }
 
 SEXP R_mongo_restore(SEXP con, SEXP ptr_col, SEXP verb) {
-  bool verbose = asLogical(verb);
+  bool verbose = Rf_asLogical(verb);
   mongoc_collection_t *col = r2col(ptr_col);
   bson_reader_t *reader = bson_reader_new_from_handle(&con, bson_reader_feed, bson_reader_finalize);
   mongoc_bulk_operation_t *bulk = NULL;
@@ -54,7 +54,7 @@ SEXP R_mongo_restore(SEXP con, SEXP ptr_col, SEXP verb) {
     if(!mongoc_bulk_operation_execute (bulk, &reply, &err)){
       bson_reader_destroy(reader);
       mongoc_bulk_operation_destroy (bulk);
-      error(err.message);
+      Rf_error(err.message);
     }
 
     if(verbose)
@@ -65,9 +65,9 @@ SEXP R_mongo_restore(SEXP con, SEXP ptr_col, SEXP verb) {
     Rprintf("\rDone! Inserted total of %d records.\n", count);
 
   if (!done)
-    warning("Failed to read all documents.\n");
+    Rf_warning("Failed to read all documents.\n");
 
   bson_reader_destroy(reader);
   mongoc_bulk_operation_destroy (bulk);
-  return ScalarInteger(count);
+  return Rf_ScalarInteger(count);
 }
